"""
This module provides utility functions for the SaltyRTC Signalling
Server.
"""
import binascii
import logging
import ssl

import libnacl
import libnacl.public

__all__ = (
    'logger_group',
    'enable_logging',
    'disable_logging',
    'get_logger',
    'consteq',
    'create_ssl_context',
    'load_permanent_key',
)


# noinspection PyUnusedLocal,PyPropertyDefinition
def _logging_error(*args, **kwargs):
    raise ImportError('Please install saltyrtc.server[logging] for logging support')


try:
    # noinspection PyPackageRequirements,PyUnresolvedReferences
    import logbook
except ImportError:
    class _Logger:
        """
        Dummy logger in case :mod:`logbook` is not present.
        """
        def __init__(self, name, level=0):
            self.name = name
            self.level = level
        debug = info = warn = warning = notice = error = exception = \
            critical = log = lambda *a, **kw: None

    class _LoggerGroup:
        """
        Dummy logger group in case :mod:`logbook` is not present.
        """
        def __init__(self, loggers=None, level=0, processor=None):
            self.loggers = loggers
            self.level = level
            self.processor = processor

        disabled = property(lambda: True, _logging_error)
        add_logger = remove_logger = process_record = _logging_error

    _logger_redirect_handler = None
    _logger_convert_level_handler = None
else:
    _Logger = logbook.Logger

    # noinspection PyPep8Naming
    def _LoggerGroup():
        group = logbook.LoggerGroup()
        group.disabled = True
        return group

    _logger_redirect_handler = logbook.compat.RedirectLoggingHandler()
    _logger_convert_level_handler = logbook.compat.LoggingHandler()


# Create logger group
logger_group = _LoggerGroup()


def _convert_level(logging_level):
    """
    Convert a :mod:`logging` level to a :mod:`logbook` level.

    Arguments:
        - `logging_level`: A :mod:`logging` level.

    Raises :class:`ImportError` in case :mod:`logbook` is not
    installed.
    """
    if _logger_convert_level_handler is None:
        _logging_error()
    return _logger_convert_level_handler.convert_level(logging_level)


def _redirect_logging_loggers(logging_loggers, remove=False):
    """
    Enable logging and redirect :mod:`logging` loggers of dependencies.

    Arguments:
        - `logging_loggers`: A dictionary containing :mod:`logging`
          logger names as key and their respective :mod:`logging` level
          as value. These loggers will be redirected to logbook.
        - `remove`: Flag to remove the redirect handler from each
          logger instead of adding it.

    Raises :class:`ImportError` in case :mod:`logbook` is not
    installed.
    """
    if _logger_redirect_handler is None:
        _logging_error()

    # At this point, logbook is either defined or an error has been returned
    for name, level in logging_loggers.items():
        # Lookup logger and translate level
        logger = logging.getLogger(name)
        logger.setLevel(_convert_level(level))

        # Add or remove redirect handler.
        if remove:
            logger.removeHandler(_logger_redirect_handler)
        else:
            logger.addHandler(_logger_redirect_handler)


def enable_logging(level=None, redirect_loggers=None):
    """
    Enable logging for the *saltyrtc* logger group.

    Arguments:
        - `level`: A :mod:`logbook` logging level. Defaults to
          ``WARNING``.
        - `redirect_loggers`: A dictionary containing :mod:`logging`
          logger names as key and their respective :mod:`logging` level
          as value. Each logger will be looked up and redirected to
          :mod:`logbook`. Defaults to an empty dictionary.

    Raises :class:`ImportError` in case :mod:`logbook` is not
    installed.
    """
    if _logger_convert_level_handler is None:
        _logging_error()

    # At this point, logbook is either defined or an error has been returned
    if level is None:
        level = logbook.WARNING
    logger_group.disabled = False
    logger_group.level = level
    if redirect_loggers is not None:
        _redirect_logging_loggers(redirect_loggers, remove=False)


def disable_logging(redirect_loggers=None):
    """
    Disable logging for the *saltyrtc* logger group.

    Arguments:
        - `level`: A :mod:`logbook` logging level.
        - `redirect_loggers`: A dictionary containing :mod:`logging`
          logger names as key and their respective :mod:`logging` level
          as value. Each logger will be looked up and removed from the
          redirect handler. Defaults to an empty dictionary.

    Raises :class:`ImportError` in case :mod:`logbook` is not
    installed.
    """
    logger_group.disabled = True
    if redirect_loggers is not None:
        _redirect_logging_loggers(redirect_loggers, remove=True)


def get_logger(name=None, level=None):
    """
    Return a :class:`logbook.Logger`.

    Arguments:
        - `name`: The name of a specific sub-logger.
        - `level`: A :mod:`logbook` logging level.
    """
    if _logger_convert_level_handler is None:
        _logging_error()

    # At this point, logbook is either defined or an error has been returned
    if level is None:
        level = logbook.NOTSET
    base_name = 'saltyrtc'
    name = base_name if name is None else '.'.join((base_name, name))

    # Create new logger and add to group
    logger = logbook.Logger(name=name, level=level)
    logger_group.add_logger(logger)
    return logger


def consteq(left, right):
    """
    Compares two byte instances with one another. If `a` and `b` have
    different lengths, return `False` immediately. Otherwise `a` and `b`
    will be compared in constant time.

    Return `True` in case `a` and `b` are equal. Otherwise `False`.

    Raises :exc:`TypeError` in case `a` and `b` are not both of the type
    :class:`bytes`.
    """
    return libnacl.bytes_eq(left, right)


def create_ssl_context(certfile, keyfile=None, dh_params_file=None):
    """
    Create and return a :class:`ssl.SSLContext` for the server.
    The settings are chosen by the :mod:`ssl` module, and usually
    represent a higher security level than when calling the
    :class:`ssl.SSLContext` constructor directly.

    Arguments:
        - `certfile`: Path to a file in PEM format containing the
          SSL certificate of the server.
        - `keyfile`: Path to a file that contains the private key.
          Will be read from `certfile` if not present.
        - `dh_params_file`: Path to a file in PEM format containing
          Diffie-Hellman parameters for DH(E) and ECDH(E). Optional
          but highly recommended.
    """
    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_context.load_cert_chain(certfile=certfile, keyfile=keyfile)
    if dh_params_file is not None:
        ssl_context.load_dh_params(dh_params_file)
    return ssl_context


def load_permanent_key(key):
    """
    Decode a hex-encoded NaCl private permanent key or read it from a
    file.

    Arguments:
        - `key`: A hex-encoded 32 bytes private permanent key or the
          name of a file which contains the key.

    Raises :exc:`ValueError` in case the key could not be found or
    is not a valid hex-encoded NaCl private key.

    Return a:class:`libnacl.public.SecretKey` instance.
    """
    # Read key file (if any)
    try:
        with open(key) as file:
            key = file.readline().strip()
    except IOError:
        pass

    # Un-hexlify
    try:
        key = binascii.unhexlify(key)
    except binascii.Error as exc:
        raise ValueError('Could not decode key') from exc

    # Convert to private key (raises ValueError on its own)
    return libnacl.public.SecretKey(sk=key)



import asyncio
import os
import struct

import libnacl
import libnacl.public
import websockets

from . import util
from .common import (
    KEEP_ALIVE_INTERVAL_DEFAULT,
    KEEP_ALIVE_INTERVAL_MIN,
    KEEP_ALIVE_TIMEOUT,
    KEY_LENGTH,
    AddressType,
    OverflowSentinel,
    available_slot_range,
    is_initiator_id,
    is_responder_id,
)
from .exception import (
    Disconnected,
    InternalError,
    MessageError,
    MessageFlowError,
    SlotsFullError,
)
from .message import unpack

__all__ = (
    'Path',
    'PathClient',
    'Protocol',
)


class Path:
    __slots__ = ('_slots', 'log', 'initiator_key', 'number')

    def __init__(self, initiator_key, number):
        self._slots = {id_: None for id_ in available_slot_range()}
        self.log = util.get_logger('path.{}'.format(number))
        self.initiator_key = initiator_key
        self.number = number

    @property
    def empty(self):
        """
        Return `True` in case the path is empty. A call to this property
        will also remove clients from the path whose connections are
        closed but have not been removed from the path. (However, in
        case that the path is not empty, this property does not ensure
        that all disconnected clients will be removed.)
        """
        for client in self._slots.values():
            if client is not None:
                if client.connection_closed.done():
                    self.remove_client(client)
                    self.log.notice('Removed dead client {}', client)
                else:
                    return False
        return True

    def get_initiator(self):
        """
        Return the initiator's :class:`PathClient` instance or `None`.
        """
        return self._slots.get(AddressType.initiator)

    def set_initiator(self, initiator):
        """
        Set the initiator's :class:`PathClient` instance.

        Arguments:
            - `initiator`: A :class:`PathClient` instance.

        Return the previously set initiator or `None`.
        """
        previous_initiator = self._slots.get(AddressType.initiator)
        self._slots[AddressType.initiator] = initiator
        self.log.debug('Set initiator {}', initiator)
        # Update initiator's log name
        initiator.update_log_name(AddressType.initiator)
        # Authenticated, assign id
        initiator.authenticated = True
        initiator.id = AddressType.initiator
        # Return previous initiator
        return previous_initiator

    def get_responder(self, id_):
        """
        Return a responder's :class:`PathClient` instance or `None`.

        Arguments:
            - `id_`: The identifier of the responder.

        Raises :exc:`ValueError` if `id_` is not a valid responder
        identifier.
        """
        if not is_responder_id(id_):
            raise ValueError('Invalid responder identifier')
        return self._slots.get(id_)

    def get_responder_ids(self):
        """
        Return a list of responder's identifiers (slots).
        """
        return [id_ for id_, responder in self._slots.items()
                if is_responder_id(id_) and responder is not None]

    def add_responder(self, responder):
        """
        Set a responder's :class:`PathClient` instance.

        Arguments:
            - `client`: A :class:`PathClient` instance.

        Raises :exc:`SlotsFullError` if no free slot exists on the path.

        Return the assigned slot identifier.
        """
        for id_, client in self._slots.items():
            if is_responder_id(id_) and client is None:
                self._slots[id_] = responder
                self.log.debug('Added responder {}', responder)
                # Update responder's log name
                responder.update_log_name(id_)
                # Authenticated, set and return assigned slot id
                responder.authenticated = True
                responder.id = id_
                return id_
        raise SlotsFullError('No free slots on path')

    def remove_client(self, client):
        """
        Remove a client (initiator or responder) from the
        :class:`Path`.

        Arguments:
             - `client`: The :class:`PathClient` instance.

        Raises :exc:`ValueError` in case the client provided an
        invalid slot identifier.
        """

        if not client.authenticated:
            # Client has not been authenticated. Nothing to do.
            return
        id_ = client.id

        # Get client instance
        try:
            slot_client = self._slots[id_]
        except KeyError:
            raise ValueError('Invalid slot identifier: {}'.format(id_))

        # Compare client instances
        if client != slot_client:
            # Note: This is absolutely fine and happens when another initiator
            #        takes the place of a previous initiator.
            return

        # Remove client from slot
        self._slots[id_] = None
        self.log.debug('Removed {}', 'initiator' if is_initiator_id(id_) else 'responder')


class PathClient:
    __slots__ = (
        '_loop',
        '_connection',
        '_client_key',
        '_server_permanent_key',
        '_server_session_key',
        '_sequence_number_out',
        '_sequence_number_in',
        '_cookie_out',
        '_cookie_in',
        '_combined_sequence_number_out',
        '_combined_sequence_number_in',
        '_box',
        '_sign_box',
        '_id',
        '_keep_alive_interval',
        'log',
        'type',
        'authenticated',
        'keep_alive_timeout',
        'keep_alive_pings',
        '_task_queue'
    )

    def __init__(
            self, connection, path_number, initiator_key,
            server_session_key=None, loop=None
    ):
        self._loop = asyncio.get_event_loop() if loop is None else loop
        self._connection = connection
        self._client_key = initiator_key
        self._server_permanent_key = None
        self._server_session_key = server_session_key
        self._cookie_out = None
        self._cookie_in = None
        self._combined_sequence_number_out = None
        self._combined_sequence_number_in = None
        self._box = None
        self._sign_box = None
        self._id = AddressType.server
        self._keep_alive_interval = KEEP_ALIVE_INTERVAL_DEFAULT
        self.log = util.get_logger('path.{}.client.{:x}'.format(path_number, id(self)))
        self.type = None
        self.authenticated = False
        self.keep_alive_timeout = KEEP_ALIVE_TIMEOUT
        self.keep_alive_pings = 0

        # Queue for tasks to be run on the client (relay messages, closing, ...)
        self._task_queue = asyncio.Queue(loop=self._loop)

    def __str__(self):
        type_ = self.type
        if type_ is None:
            type_ = 'undetermined'
        return 'PathClient(role=0x{:02x}, id={}, at={})'.format(
            type_, self._id, hex(id(self)))

    @property
    def connection_closed(self):
        """
        Return the 'connection_closed' future of the underlying
        WebSocket connection.
        """
        return self._connection.connection_closed

    @property
    def id(self):
        """
        Return the assigned id on the :class:`Path`.
        """
        return self._id

    @id.setter
    def id(self, id_):
        """
        Assign the id. Only :class:`Path` may set the id!
        """
        self._id = id_
        self.log.debug('Assigned id: {}', id_)

    @property
    def keep_alive_interval(self):
        """
        Return the currently set keep alive interval.
        """
        return self._keep_alive_interval

    @keep_alive_interval.setter
    def keep_alive_interval(self, interval):
        """
        Assign a new keep alive interval. Will ignore values less than
        `KEEP_ALIVE_INTERVAL_MIN`.
        """
        if interval >= KEEP_ALIVE_INTERVAL_MIN:
            self._keep_alive_interval = interval

    @property
    def client_key(self):
        """
        Return the client's permanent key as :class:`bytes`.

        .. warning:: This is the initiator's key at default if no other
                     client key has been set!
        """
        return self._client_key

    @property
    def server_key(self):
        """
        Return the server's session :class:`libnacl.public.SecretKey`
        instance.
        """
        if self._server_session_key is None:
            self._server_session_key = libnacl.public.SecretKey()
        return self._server_session_key

    @property
    def server_permanent_key(self):
        """
        Return the server's permanent :class:`libnacl.public.SecretKey`
        instance chosen by the client.

        Raises `InternalError` in case the key has not been set, yet.
        """
        if self._server_permanent_key is None:
            raise InternalError("Server's permanent secret key instance not set")
        return self._server_permanent_key

    @server_permanent_key.setter
    def server_permanent_key(self, key):
        """
        Set the server's permanent :class:`libnacl.public.SecretKey`
        instance chosen by the client.
        """
        self._server_permanent_key = key

    @property
    def box(self):
        """
        Return the session's :class:`libnacl.public.Box` instance.
        """
        if self._box is None:
            self._box = libnacl.public.Box(self.server_key, self._client_key)
        return self._box

    @property
    def sign_box(self):
        """
        Return the :class:`libnacl.public.Box` instance that is used for
        signing the keys in the 'server-auth' message.

        Raises `InternalError` in case the server's permanent key has
        not been set, yet.
        """
        if self._sign_box is None:
            self._sign_box = libnacl.public.Box(
                self.server_permanent_key, self._client_key)
        return self._sign_box

    @property
    def cookie_out(self):
        """
        Return the cookie of the server (outgoing messages).
        """
        if self._cookie_out is None:
            self._cookie_out = os.urandom(16)
        return self._cookie_out

    @property
    def cookie_in(self):
        """
        Return the cookie of the client (incoming messages).
        """
        return self._cookie_in

    @property
    def combined_sequence_number_out(self):
        """
        Return the pending combined sequence number of the server
        (outgoing messages).
        """
        if self._combined_sequence_number_out is None:
            # Initialise the trailing 32 bits of the uint48 number with random bits
            initial_number, *_ = struct.unpack('!Q', b'\x00' * 4 + os.urandom(4))
            self._combined_sequence_number_out = initial_number
        return self._combined_sequence_number_out

    @combined_sequence_number_out.setter
    def combined_sequence_number_out(self, combined_sequence_number_out):
        """
        Update the combined sequence number of the server (outgoing
        messages). Will set the number to OverflowSentinel in case
        the number would overflow a 48-bit unsigned integer.
        """
        csn = self._validate_combined_sequence_number(combined_sequence_number_out)
        self._combined_sequence_number_out = csn

    @property
    def combined_sequence_number_in(self):
        """
        Return the pending combined sequence number of the client
        (incoming messages).
        """
        return self._combined_sequence_number_in

    @combined_sequence_number_in.setter
    def combined_sequence_number_in(self, combined_sequence_number_in):
        """
        Update the combined sequence number of the client (incoming
        messages). Will set the number to OverflowSentinel in case
        the number would overflow a 48-bit unsigned integer.
        """
        csn = self._validate_combined_sequence_number(combined_sequence_number_in)
        self._combined_sequence_number_in = csn

    @staticmethod
    def _validate_combined_sequence_number(combined_sequence_number):
        """
        Validate a combined sequence number.

        Return OverflowSentinel in case the number would overflow a
        48-bit unsigned integer, otherwise the passed sequence number.
        """
        if combined_sequence_number & 0xf000000000000 != 0:
            return OverflowSentinel
        else:
            return combined_sequence_number

    def set_client_key(self, public_key):
        """
        Set the public key of the client and update the internal box.

        Arguments:
            - `public_key`: A :class:`libnacl.public.PublicKey`.
        """
        self._client_key = public_key
        self._box = libnacl.public.Box(self.server_key, public_key)
        self.log.debug('Client key updated')

    def update_log_name(self, slot_id):
        """
        Update the logger's name by the assigned slot identifier.

        Arguments:
            - `slot_id`: The slot identifier of the client.
        """
        self.log.name += '.0x{:02x}'.format(slot_id)

    def valid_cookie(self, cookie_in):
        """
        Return `True` if the 16 byte cookie is the valid cookie of the
        client (or the cookie has not been set, yet).
        """
        if self.cookie_in is None:
            # Ensure that the client uses another cookie than we do
            if cookie_in == self.cookie_out:
                self.log.notice('Server and client cookies are the same')
                return False

            # First message: Set cookie
            self._cookie_in = cookie_in
            return True
        else:
            if cookie_in != self.cookie_in:
                self.log.notice('Client sent wrong cookie')
                return False
            return True

    def validate_combined_sequence_number(self, combined_sequence_number_in):
        """
        Set or validate the combined sequence number for incoming
        messages from the client.

        Raises:
            - :exc:`MessageError` in case this is the first message
              from the client and the leading 16 bits of the combined
              sequence number are not `0`.
            - :exc:`MessageError` in case the 6 byte combined sequence
              number is not the expected sequence number of the client.
            - :exc:`MessageFlowError` in case no messages can be
              received from the client as the combined sequence number
              has reset to `0`.
        """
        if self.combined_sequence_number_in is None:
            # Ensure that the leading 16 bits are 0
            if combined_sequence_number_in & 0xffff00000000 != 0:
                raise MessageError('Invalid sequence number, leading 16 bits are not 0')

            # First message: Set combined sequence number
            self._combined_sequence_number_in = combined_sequence_number_in

        # Ensure that the incoming CSN counter did not overflow
        if self.combined_sequence_number_in == OverflowSentinel:
            raise MessageFlowError(('Cannot receive any more messages, due to a '
                                    'sequence number counter overflow'))

        # Check that the CSN matches the expected CSN
        if combined_sequence_number_in != self.combined_sequence_number_in:
            raise MessageError('Invalid sequence number, expected {}, got {}'.format(
                self.combined_sequence_number_in, combined_sequence_number_in
            ))

    def p2p_allowed(self, destination_type):
        """
        Return `True` if :class:`RawMessage` instances are allowed and
        can be sent to the requested :class:`AddressType`.
        """
        return self.authenticated and self.type != destination_type

    @asyncio.coroutine
    def enqueue_task(self, coroutine_or_task):
        """
        Enqueue a coroutine or task into the task queue of the
        client.

        Arguments:
            - `coroutine_or_task`: A coroutine or a
              :class:`asyncio.Task`.
        """
        yield from self._task_queue.put(coroutine_or_task)

    @asyncio.coroutine
    def dequeue_task(self):
        """
        Dequeue and return a coroutine or task from the task queue of
        the client.

        Shall only be called from the client's :class:`Protocol`
        instance.
        """
        return (yield from self._task_queue.get())

    @asyncio.coroutine
    def send(self, message):
        """
        Disconnected
        MessageError
        MessageFlowError
        """
        # Pack
        self.log.debug('Packing message: {}', message.type)
        data = message.pack(self)
        self.log.trace('server >> {}', message)

        # Send data
        self.log.debug('Sending message')
        try:
            yield from self._connection.send(data)
        except websockets.ConnectionClosed as exc:
            self.log.debug('Connection closed while sending')
            raise Disconnected(exc.code) from exc

    @asyncio.coroutine
    def receive(self):
        """
        Disconnected
        """
        # Receive data
        try:
            data = yield from self._connection.recv()
        except websockets.ConnectionClosed as exc:
            self.log.debug('Connection closed while receiving')
            raise Disconnected(exc.code) from exc
        self.log.debug('Received message')

        # Unpack data and return
        message = unpack(self, data)
        self.log.debug('Unpacked message: {}', message.type)
        self.log.trace('server << {}', message)
        return message

    @asyncio.coroutine
    def ping(self):
        """
        Disconnected
        """
        self.log.debug('Sending ping')
        try:
            return (yield from self._connection.ping())
        except websockets.ConnectionClosed as exc:
            self.log.debug('Connection closed while pinging')
            raise Disconnected(exc.code) from exc

    @asyncio.coroutine
    def close(self, code=1000):
        # Note: We are not sending a reason for security reasons.
        yield from self._connection.close(code=code)


class Protocol:
    PATH_LENGTH = KEY_LENGTH * 2



import asyncio
import functools
import os
import socket
import ssl
import struct
import subprocess
import sys
from contextlib import closing

import libnacl.public
import logbook
import pytest
import umsgpack
import websockets

from saltyrtc.server import (
    NONCE_FORMATTER,
    NONCE_LENGTH,
    Event,
    Server,
    SubProtocol,
    serve,
    util,
)


class CalledProcessError(subprocess.CalledProcessError):
    def __str__(self):
        return "Command '{}' returned non-zero exit status {}:\n{}".format(
            self.cmd, self.returncode, self.output)


def pytest_addoption(parser):
    # 'loop' parameter
    help_ = 'Use a different event loop, supported: asyncio, uvloop'
    parser.addoption('--loop', action='store', help=help_)

    # 'timeout' parameter
    help_ = 'Use a specific timeout in seconds (float) for tests'
    parser.addoption('--timeout', action='store', help=help_)


def pytest_report_header(config):
    lines = [
        'Using event loop: {}'.format(default_event_loop(config=config)),
        'Using timeout: {}s'.format(_get_timeout(config=config)),
    ]
    return '\n'.join(lines)


def pytest_namespace():
    try:
        # noinspection PyPackageRequirements,PyUnresolvedReferences
        import uvloop  # noqa
        have_uvloop = True
    except ImportError:
        have_uvloop = False
    saltyrtc = {
        'have_uvloop': pytest.mark.skipif(not have_uvloop, reason='requires uvloop'),
        'no_uvloop': pytest.mark.skipif(
            have_uvloop, reason='requires uvloop to be not installed'),
        'ip': '127.0.0.1',
        'port': 8766,
        'cli_path': os.path.join(sys.exec_prefix, 'bin', 'saltyrtc-server'),
        'cert': os.path.normpath(
            os.path.join(os.path.abspath(__file__), os.pardir, 'cert.pem')),
        'dh_params': os.path.normpath(
            os.path.join(os.path.abspath(__file__), os.pardir, 'dh2048.pem')),
        'permanent_key_primary': os.path.normpath(
            os.path.join(os.path.abspath(__file__), os.pardir, 'permanent.key')),
        'permanent_key_secondary':
            'b452e8a5abf54c5258db323b88d03cb9e002a4a84ba6f37715678901c20411c7',
        'subprotocols': [
            SubProtocol.saltyrtc_v1.value
        ],
        'timeout': 0.4,
        'run_long_tests': False,
    }
    saltyrtc['long_test'] = pytest.mark.skipif(
        not saltyrtc['run_long_tests'],
        reason='requires explicitly enabled long tests')
    return {'saltyrtc': saltyrtc}


def default_event_loop(request=None, config=None):
    if request is not None:
        config = request.config
    loop = config.getoption("--loop")
    if loop == 'uvloop':
        import uvloop
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    else:
        loop = 'asyncio'
    return loop


def unused_tcp_port():
    """
    Find an unused localhost TCP port from 1024-65535 and return it.
    """
    with closing(socket.socket()) as sock:
        sock.bind((pytest.saltyrtc.ip, 0))
        return sock.getsockname()[1]


def url(host, port):
    """
    Return the URL where the server can be reached.
    """
    return 'wss://{}:{}'.format(host, port)


def key_pair():
    """
    Return a NaCl key pair.
    """
    return libnacl.public.SecretKey()


def key_path(key_pair):
    """
    Return the hexadecimal key path from a key pair using the public
    key.

    Arguments:
        - `key_pair`: A :class:`libnacl.public.SecretKey` instance.
    """
    return key_pair.hex_pk().decode()


def random_cookie():
    """
    Return a random cookie for the client.
    """
    return os.urandom(16)


def _get_timeout(timeout=None, request=None, config=None):
    """
    Return the defined timeout.
    """
    if timeout is None:
        timeout = pytest.saltyrtc.timeout
    if request is not None:
        config = request.config
    option_timeout = config.getoption("--timeout")
    if option_timeout is not None:
        return max(timeout, float(option_timeout))
    else:
        return timeout


def _sleep(**kwargs):
    """
    Sleep *timeout* seconds.
    """
    @asyncio.coroutine
    def __sleep(timeout=None):
        yield from asyncio.sleep(_get_timeout(timeout=timeout, **kwargs))

    return __sleep


@pytest.fixture(scope='module')
def event_loop(request):
    """
    Create an instance of the requested event loop.
    """
    default_event_loop(request=request)

    # Close previous event loop
    policy = asyncio.get_event_loop_policy()
    policy.get_event_loop().close()

    # Create new event loop
    _event_loop = policy.new_event_loop()
    policy.set_event_loop(_event_loop)

    def fin():
        _event_loop.close()

    # Add finaliser and return new event loop
    request.addfinalizer(fin)
    return _event_loop


@pytest.fixture(scope='module')
def server_permanent_keys():
    """
    Return the server's permanent test NaCl key pairs.
    """
    return [
        util.load_permanent_key(pytest.saltyrtc.permanent_key_primary),
        util.load_permanent_key(pytest.saltyrtc.permanent_key_secondary)
    ]


@pytest.fixture(scope='module')
def server_key():
    """
    Return a server NaCl key pair to be used by the server only.
    """
    return key_pair()


@pytest.fixture(scope='module')
def initiator_key():
    """
    Return a client NaCl key pair to be used by the initiator only.
    """
    return key_pair()


@pytest.fixture(scope='module')
def responder_key():
    """
    Return a client NaCl key pair to be used by the responder only.
    """
    return key_pair()


@pytest.fixture(scope='module')
def sleep(request):
    """
    Sleep *timeout* seconds.
    """
    return _sleep(request=request)


@pytest.fixture(scope='module')
def cookie_factory():
    """
    A cookie factory for random cookies.
    """
    return random_cookie


@pytest.fixture(scope='module')
def url_factory(server):
    """
    Return the URL where the server can be reached.
    """
    server_ = server

    def _url_factory(server=None):
        if server is None:
            server = server_
        return url(*server.address)
    return _url_factory


class TestServer(Server):
    def __init__(self, *args, timeout=None, **kwargs):
        super().__init__(*args, **kwargs)

        # Store timeout
        self.timeout = timeout

        # Newly connection closed future
        self._new_connection_closed_future = asyncio.Future(loop=self._loop)

    def raise_event(self, event: Event, *data):
        super().raise_event(event, *data)
        if event == Event.disconnected:
            self._new_connection_closed_future.set_result(None)
            self._new_connection_closed_future = asyncio.Future(loop=self._loop)

    @asyncio.coroutine
    def wait_connections_closed(self):
        self._log.debug('#protocols remaining: {}', len(self.protocols))
        return asyncio.wait_for(
            self._wait_connections_closed(), timeout=self.timeout, loop=self._loop)

    def new_connection_closed(self, future=None):
        if future is None:
            future = self._new_connection_closed_future
        return asyncio.wait_for(
            future, timeout=self.timeout, loop=self._loop)

    def new_connection_closed_delayed(self):
        return functools.partial(
            self.new_connection_closed, future=self._new_connection_closed_future)


@pytest.fixture(scope='module')
def server_factory(request, event_loop, server_permanent_keys):
    """
    Return a factory to create :class:`saltyrtc.Server` instances.
    """
    # Enable asyncio debug logging
    os.environ['PYTHONASYNCIODEBUG'] = '1'

    # Enable logging
    util.enable_logging(level=logbook.NOTICE, redirect_loggers={
        'asyncio': logbook.WARNING,
        'websockets': logbook.WARNING,
    })

    # Push handler
    logging_handler = logbook.StderrHandler()
    logging_handler.push_application()

    _server_instances = []

    def _server_factory(permanent_keys=None):
        if permanent_keys is None:
            permanent_keys = server_permanent_keys

        # Setup server
        port = unused_tcp_port()
        coroutine = serve(
            util.create_ssl_context(
                pytest.saltyrtc.cert, dh_params_file=pytest.saltyrtc.dh_params),
            permanent_keys,
            host=pytest.saltyrtc.ip,
            port=port,
            loop=event_loop,
            server_class=TestServer,
        )
        server_ = event_loop.run_until_complete(coroutine)
        # Inject timeout and address (little bit of a hack but meh...)
        server_.timeout = _get_timeout(request=request)
        server_.address = (pytest.saltyrtc.ip, port)

        _server_instances.append(server_)

        def fin():
            server_.close()
            event_loop.run_until_complete(server_.wait_closed())
            _server_instances.remove(server_)
            if len(_server_instances) == 0:
                logging_handler.pop_application()

        request.addfinalizer(fin)
        return server_
    return _server_factory


@pytest.fixture(scope='module')
def server(server_factory):
    """
    Return a :class:`saltyrtc.Server` instance.
    """
    return server_factory()


@pytest.fixture(scope='module')
def server_no_key(server_factory):
    """
    Return a :class:`saltyrtc.Server` instance that has no permanent
    key pair.
    """
    return server_factory(permanent_keys=[])


class _DefaultBox:
    pass


class Client:
    def __init__(self, ws_client, pack_message, unpack_message, request, timeout=None):
        self.ws_client = ws_client
        self.pack_and_send = pack_message
        self.recv_and_unpack = unpack_message
        self.timeout = _get_timeout(timeout=timeout, request=request)
        self.session_key = None
        self.box = None

    def send(self, nonce, message, box=_DefaultBox, timeout=None, pack=True):
        if timeout is None:
            timeout = self.timeout
        return (yield from self.pack_and_send(
            self.ws_client, nonce, message,
            box=self.box if box == _DefaultBox else box, timeout=timeout, pack=pack
        ))

    def recv(self, box=_DefaultBox, timeout=None):
        if timeout is None:
            timeout = self.timeout
        return (yield from self.recv_and_unpack(
            self.ws_client,
            box=self.box if box == _DefaultBox else box, timeout=timeout
        ))

    def close(self):
        return self.ws_client.close()


@pytest.fixture(scope='module')
def ws_client_factory(initiator_key, event_loop, server):
    """
    Return a simplified :class:`websockets.client.connect` wrapper
    where no parameters are required.
    """
    # Note: The `server` argument is only required to fire up the server.
    server_ = server

    # Create SSL context
    ssl_context = ssl.create_default_context(
        ssl.Purpose.SERVER_AUTH, cafile=pytest.saltyrtc.cert)
    ssl_context.load_dh_params(pytest.saltyrtc.dh_params)

    def _ws_client_factory(server=None, path=None, **kwargs):
        if server is None:
            server = server_
        if path is None:
            path = '{}/{}'.format(url(*server.address), key_path(initiator_key))
        _kwargs = {
            'subprotocols': pytest.saltyrtc.subprotocols,
            'ssl': ssl_context,
            'loop': event_loop,
        }
        _kwargs.update(kwargs)
        return websockets.connect(path, **_kwargs)
    return _ws_client_factory


@pytest.fixture(scope='module')
def client_factory(
        request, initiator_key, event_loop, server, server_permanent_keys, responder_key,
        pack_nonce, pack_message, unpack_message
):
    """
    Return a simplified :class:`websockets.client.connect` wrapper
    where no parameters are required.
    """
    # Note: The `server` argument is only required to fire up the server.
    server_ = server

    # Create SSL context
    ssl_context = ssl.create_default_context(
        ssl.Purpose.SERVER_AUTH, cafile=pytest.saltyrtc.cert)
    ssl_context.load_dh_params(pytest.saltyrtc.dh_params)

    @asyncio.coroutine
    def _client_factory(
            server=None, ws_client=None,
            path=initiator_key, timeout=None, csn=None, cookie=None, permanent_key=None,
            ping_interval=None, explicit_permanent_key=False,
            initiator_handshake=False, responder_handshake=False,
            **kwargs
    ):
        if server is None:
            server = server_
        if cookie is None:
            cookie = random_cookie()
        if permanent_key is None:
            permanent_key = server_permanent_keys[0].pk
        _kwargs = {
            'subprotocols': pytest.saltyrtc.subprotocols,
            'ssl': ssl_context,
            'loop': event_loop,
        }
        _kwargs.update(kwargs)
        if ws_client is None:
            ws_client = yield from websockets.connect(
                '{}/{}'.format(url(*server.address), key_path(path)),
                **_kwargs
            )
        client = Client(
            ws_client, pack_message, unpack_message,
            request, timeout=timeout
        )
        nonces = {}

        if not initiator_handshake and not responder_handshake:
            return client

        # Do handshake
        key = initiator_key if initiator_handshake else responder_key

        # server-hello
        message, nonce, sck, s, d, start_scsn = yield from client.recv()
        ssk = message['key']
        nonces['server-hello'] = nonce

        cck, ccsn = cookie, 2 ** 32 - 1 if csn is None else csn
        if responder_handshake:
            # client-hello
            nonce = pack_nonce(cck, 0x00, 0x00, ccsn)
            nonces['client-hello'] = nonce
            yield from client.send(nonce, {
                'type': 'client-hello',
                'key': responder_key.pk,
            })
            ccsn += 1

        # client-auth
        client.box = libnacl.public.Box(sk=key, pk=ssk)
        nonce = pack_nonce(cck, 0x00, 0x00, ccsn)
        nonces['client-auth'] = nonce
        payload = {
            'type': 'client-auth',
            'your_cookie': sck,
            'subprotocols': pytest.saltyrtc.subprotocols,
        }
        if ping_interval is not None:
            payload['ping_interval'] = ping_interval
        if explicit_permanent_key is not None:
            payload['your_key'] = permanent_key
        yield from client.send(nonce, payload)
        ccsn += 1

        # server-auth
        client.sign_box = libnacl.public.Box(sk=key, pk=permanent_key)
        message, nonce, ck, s, d, scsn = yield from client.recv()
        nonces['server-auth'] = nonce

        # Return client and additional data
        additional_data = {
            'id': d,
            'sck': sck,
            'start_scsn': start_scsn,
            'cck': cck,
            'ccsn': ccsn,
            'ssk': ssk,
            'nonces': nonces,
            'signed_keys': message['signed_keys']
        }
        if initiator_handshake:
            additional_data['responders'] = message['responders']
        else:
            additional_data['initiator_connected'] = message['initiator_connected']
        return client, additional_data
    return _client_factory


@pytest.fixture(scope='module')
def unpack_message(request, event_loop):
    @asyncio.coroutine
    def _unpack_message(client, box=None, timeout=None):
        timeout = _get_timeout(timeout=timeout, request=request)
        data = yield from asyncio.wait_for(client.recv(), timeout, loop=event_loop)
        nonce = data[:NONCE_LENGTH]
        (cookie,
         source, destination,
         combined_sequence_number) = struct.unpack(NONCE_FORMATTER, nonce)
        combined_sequence_number, *_ = struct.unpack(
            '!Q', b'\x00\x00' + combined_sequence_number)
        data = data[NONCE_LENGTH:]
        if box is not None:
            data = box.decrypt(data, nonce=nonce)
        else:
            nonce = None
        message = umsgpack.unpackb(data)
        return (
            message,
            nonce,
            cookie,
            source, destination,
            combined_sequence_number
        )
    return _unpack_message


@pytest.fixture(scope='module')
def pack_nonce():
    def _pack_nonce(cookie, source, destination, combined_sequence_number):
        return struct.pack(
            NONCE_FORMATTER,
            cookie,
            source, destination,
            struct.pack('!Q', combined_sequence_number)[2:]
        )
    return _pack_nonce


@pytest.fixture(scope='module')
def pack_message(request, event_loop):
    @asyncio.coroutine
    def _pack_message(client, nonce, message, box=None, timeout=None, pack=True):
        if pack:
            data = umsgpack.packb(message)
        else:
            data = message
        if box is not None:
            _, data = box.encrypt(data, nonce=nonce, pack_nonce=False)
        data = b''.join((nonce, data))
        timeout = _get_timeout(timeout=timeout, request=request)
        yield from asyncio.wait_for(client.send(data), timeout, loop=event_loop)
        return data
    return _pack_message


@pytest.fixture(scope='module')
def cli(request, event_loop):
    @asyncio.coroutine
    def _call_cli(*args, input=None, timeout=None, signal=None, env=None):
        # Get timeout
        timeout = _get_timeout(timeout=timeout, request=request)

        # Prepare environment
        if env is None:
            env = os.environ.copy()

        # Call CLI in subprocess and get output
        parameters = [sys.executable, pytest.saltyrtc.cli_path] + list(args)
        if isinstance(input, str):
            input = input.encode('utf-8')

        # Create process
        create = asyncio.create_subprocess_exec(
            *parameters, env=env, stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT)
        process = yield from create

        # Wait for process to terminate
        task = event_loop.create_task(process.communicate(input=input))
        maybe_shielded_task = task
        if signal is not None:
            maybe_shielded_task = asyncio.shield(maybe_shielded_task, loop=event_loop)
        output = None
        timeout_exc = False
        try:
            output, _ = yield from asyncio.wait_for(
                maybe_shielded_task, timeout, loop=event_loop)
        except asyncio.TimeoutError:
            if signal is None:
                raise
            timeout_exc = True

        # Send signal (if requested)
        if timeout_exc and signal is not None:
            try:
                signals = list(signal)
            except TypeError:
                signals = [signal]
            length = len(signals)
            for i, signal in enumerate(signals):
                shielded_task = asyncio.shield(task, loop=event_loop)
                process.send_signal(signal)
                try:
                    output, _ = yield from asyncio.wait_for(
                        shielded_task, timeout, loop=event_loop)
                except asyncio.TimeoutError:
                    if i == (length - 1):
                        task.cancel()
                        raise

        # Process output
        output = output.decode('utf-8')

        # Strip leading empty lines and pydev debugger output
        rubbish = [
            'pydev debugger: process',
            'Traceback (most recent call last):',
        ]
        lines = []
        skip_following_empty_lines = True
        for line in output.splitlines(keepends=True):
            if any((line.startswith(s) for s in rubbish)):
                skip_following_empty_lines = True
            elif not skip_following_empty_lines or len(line.strip()) > 0:
                lines.append(line)
                skip_following_empty_lines = False

        # Strip trailing empty lines
        empty_lines_count = 0
        for line in reversed(lines):
            if len(line.strip()) > 0:
                break
            empty_lines_count += 1
        if empty_lines_count > 0:
            lines = lines[:-empty_lines_count]
        output = ''.join(lines)

        # Check return code
        if process.returncode != 0:
            raise CalledProcessError(
                process.returncode, parameters, output=output)
        return output
    return _call_cli


@pytest.fixture(scope='function')
def fake_logbook_env(tmpdir):
    tmpdir.join("logbook.py").write("raise ImportError('h3h3')")
    env = os.environ.copy()
    env['PYTHONPATH'] = ':'.join((str(tmpdir), env.get('PYTHONPATH', '')))
    return env



"""
The tests provided in this module make sure that the server
instance behaves as expected.
"""

import pytest

from saltyrtc import server


class TestServer:
    @pytest.mark.asyncio
    def test_repeated_permanent_keys(self, server_permanent_keys):
        """
        Ensure the server does not accept repeated keys.
        """
        keys = server_permanent_keys + [server_permanent_keys[1]]
        with pytest.raises(server.ServerKeyError) as exc_info:
            yield from server.serve(None, keys)
        assert 'Repeated permanent keys' in str(exc_info.value)



import abc
import binascii
import io
import struct

# noinspection PyPackageRequirements
import umsgpack

from .common import sign_keys as sign_keys_
from .common import (
    COOKIE_LENGTH,
    DATA_LENGTH_MIN,
    NONCE_FORMATTER,
    NONCE_LENGTH,
    AddressType,
    CloseCode,
    MessageType,
    OverflowSentinel,
    validate_cookie,
    validate_drop_reason,
    validate_hash,
    validate_initiator_connected,
    validate_ping_interval,
    validate_public_key,
    validate_responder_id,
    validate_responder_ids,
    validate_signed_keys,
    validate_subprotocols,
)
from .exception import (
    MessageError,
    MessageFlowError,
)

__all__ = (
    'unpack',
    'AbstractMessage',
    'AbstractBaseMessage',
    'RawMessage',
    'ServerHelloMessage',
    'ClientHelloMessage',
    'ClientAuthMessage',
    'ServerAuthMessage',
    'NewInitiatorMessage',
    'NewResponderMessage',
    'DropResponderMessage',
    'SendErrorMessage',
)


def unpack(client, data):
    """
    MessageError
    MessageFlowError
    """
    return AbstractBaseMessage.unpack(client, data)


def _message_representation(class_name, nonce, payload, encrypted=None):
    hex_cookie_length = COOKIE_LENGTH * 2
    nonce_as_hex = binascii.hexlify(nonce).decode('ascii')
    nonce_as_hex = '|'.join((
        nonce_as_hex[:hex_cookie_length],
        nonce_as_hex[hex_cookie_length:hex_cookie_length + 2],
        nonce_as_hex[hex_cookie_length + 2:hex_cookie_length + 4],
        nonce_as_hex[hex_cookie_length + 4:hex_cookie_length + 8],
        nonce_as_hex[hex_cookie_length + 8:]
    ))
    if isinstance(encrypted, bool):
        encrypted_str = 'encrypted={}, '.format(encrypted)
    else:
        encrypted_str = ''
    return '{}({}nonce={}, data={})'.format(
        class_name, encrypted_str, nonce_as_hex, payload)


class AbstractMessage(metaclass=abc.ABCMeta):
    type = None

    def __init__(self, source, destination, source_type=None, destination_type=None):
        if source_type is None:
            AddressType.from_address(source)
        if destination_type is None:
            AddressType.from_address(destination)
        self.source = source
        self.destination = destination
        self.source_type = source_type
        self.destination_type = destination_type
        self._nonce = None

    @abc.abstractmethod
    def pack(self, client):
        """
        MessageError
        MessageFlowError
        """
        raise NotImplementedError

    @abc.abstractmethod
    def prepare_payload(self, client, nonce):
        """
        This method will be called as soon as the nonce has been packed
        to make late changes to the payload based on the nonce.

        Only :exc:`MessageError` may be raised.
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def unpack(cls, client, data):
        """
         MessageError
         MessageFlowError
         """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def check_payload(cls, client, payload):
        """
        MessageError
        """
        raise NotImplementedError


# noinspection PyAbstractClass
class AbstractBaseMessage(AbstractMessage, metaclass=abc.ABCMeta):
    encrypted = None

    def __new__(cls, payload, *args, **kwargs):
        # Ensure the class has implemented a class-level `type` attribute
        if cls.type not in MessageType:
            message = 'Cannot instantiate class {} with invalid message type: {}'
            raise TypeError(message.format(cls.__name__, cls.type))

        # Ensure the class has implemented a class-level `encrypted` flag
        if cls.encrypted is not True and cls.encrypted is not False:
            message = 'Cannot instantiate class {} with invalid encrypted flag: {}'
            raise TypeError(message.format(cls.__name__, cls.encrypted))

        return super().__new__(cls)

    def __init__(
            self, source, destination, payload,
            source_type=None, destination_type=None, extra=None
    ):
        super().__init__(
            source, destination,
            source_type=source_type, destination_type=destination_type
        )
        self.payload = {} if payload is None else payload
        self.extra = {} if extra is None else extra

    def __str__(self):
        return _message_representation(
            self.__class__.__name__, self._nonce, self.payload, encrypted=self.encrypted)

    @classmethod
    def _get_message_classes(cls):
        if getattr(cls, '_message_classes', None) is None:
            cls._message_classes = {
                MessageType.server_hello: ServerHelloMessage,
                MessageType.client_hello: ClientHelloMessage,
                MessageType.client_auth: ClientAuthMessage,
                MessageType.server_auth: ServerAuthMessage,
                MessageType.new_responder: NewResponderMessage,
                MessageType.drop_responder: DropResponderMessage,
                MessageType.send_error: SendErrorMessage,
            }
        return cls._message_classes

    def pack(self, client):
        """
        MessageError
        MessageFlowError
        """
        data = io.BytesIO()

        # Pack nonce
        nonce = self._pack_nonce(client)
        self._nonce = nonce  # Stored for str representation
        data.write(nonce)

        # Prepare payload
        self.prepare_payload(client, nonce)

        # Pack payload
        payload = self._pack_payload()

        # Encrypt payload if required
        if self.encrypted:
            if not client.authenticated:
                raise MessageFlowError('Cannot encrypt payload, no box available')
            payload = self._encrypt_payload(client, nonce, payload)

        # Append payload and return as bytes
        data.write(payload)
        return data.getvalue()

    def prepare_payload(self, client, nonce):
        return

    @classmethod
    def unpack(cls, client, data):
        """
        MessageError
        MessageFlowError
        """
        # Check length
        length = len(data)
        if length < DATA_LENGTH_MIN:
            raise MessageError('Message too short: {} bytes'.format(length))

        # Unpack and check nonce
        (nonce,
         source, source_type,
         destination, destination_type) = cls._unpack_nonce(data, client)

        # Decrypt if directed at us and keys have been exchanged
        # or just return a raw message to be sent to another client
        if destination_type == AddressType.server:
            data = data[NONCE_LENGTH:]
            if not client.authenticated and client.type is None:
                payload = None

                # Try client-hello (unencrypted)
                try:
                    payload = cls._unpack_payload(data)
                except MessageError:
                    pass

                # Try client-auth (encrypted)
                try:
                    payload = cls._unpack_payload(
                        cls._decrypt_payload(client, nonce, data))
                except MessageError:
                    pass

                # Still no payload?
                if payload is None:
                    message = 'Expected either client-hello or client-auth, got neither'
                    raise MessageError(message)
            else:
                # Decrypt and unpack payload
                payload = cls._unpack_payload(
                    cls._decrypt_payload(client, nonce, data))
        else:
            message = RawMessage(
                source, destination, data,
                source_type=source_type, destination_type=destination_type
            )
            message._nonce = nonce  # Stored for str representation
            return message

        # Unpack type
        try:
            type_ = payload.get('type')
        except AttributeError as exc:
            error = 'Message does not contain a dictionary: {}'
            raise MessageError(error.format(type(payload))) from exc
        try:
            type_ = MessageType(type_)
        except ValueError as exc:
            raise MessageError('Unknown message type: {}'.format(type_)) from exc

        # Check and convert payload on appropriate message class
        try:
            message_class = cls._get_message_classes()[type_]
        except KeyError as exc:
            error_message = 'Can not handle valid message type: {}'
            raise MessageError(error_message.format(type_)) from exc
        payload = message_class.check_payload(client, payload)

        # Return instance
        message = message_class(
            source, destination, payload,
            source_type=source_type, destination_type=destination_type
        )
        message._nonce = nonce  # Stored for str representation
        return message

    def _pack_nonce(self, client):
        """
        .. note:: The CSN check and incrementation can only reside here
                  because an error while packing a message will
                  create a protocol error in any case.

        MessageError
        MessageFlowError
        """
        # Ensure that the outgoing combined sequence number counter did not overflow
        if client.combined_sequence_number_out == OverflowSentinel:
            raise MessageFlowError(('Cannot send any more messages, due to a sequence '
                                    'number counter overflow'))

        # Pack nonce
        try:
            nonce = struct.pack(
                NONCE_FORMATTER,
                client.cookie_out,
                self.source, self.destination,
                struct.pack('!Q', client.combined_sequence_number_out)[2:]
            )
        except struct.error as exc:
            raise MessageError('Could not pack nonce') from exc

        # Increase outgoing combined sequence number counter
        client.combined_sequence_number_out += 1
        return nonce

    @classmethod
    def _unpack_nonce(cls, data, client):
        """
        .. note:: The CSN check and incrementation can only reside here
                  because an error while unpacking a message will
                  create a protocol error in any case.

        MessageError
        MessageFlowError
        """
        nonce = data[:NONCE_LENGTH]
        try:
            (cookie_in,
             source, destination,
             combined_sequence_number_in) = struct.unpack(NONCE_FORMATTER, nonce)
            combined_sequence_number_in, *_ = struct.unpack(
                '!Q', b'\x00\x00' + combined_sequence_number_in)
        except struct.error as exc:
            raise MessageError('Could not unpack nonce') from exc

        # Get source and destination address type
        source_type = AddressType.from_address(source)
        destination_type = AddressType.from_address(destination)

        # Validate destination
        # (Is the client allowed to send messages to the address type?)
        is_to_server = destination_type == AddressType.server
        if not is_to_server and not client.p2p_allowed(destination_type):
            error = 'Not allowed to relay messages to 0x{:02x}'
            raise MessageFlowError(error.format(destination_type))

        # Validate source
        if source != client.id:
            error_message = 'Identities do not match, expected 0x{:02x}, got 0x{:02x}'
            raise MessageError(error_message.format(client.id, source))

        # Validate cookie
        if is_to_server and not client.valid_cookie(cookie_in):
            raise MessageError('Invalid cookie: {}'.format(cookie_in))

        # Validate and increase combined sequence number
        if is_to_server:
            client.validate_combined_sequence_number(combined_sequence_number_in)
            client.combined_sequence_number_in += 1

        return nonce, source, source_type, destination, destination_type

    def _pack_payload(self):
        try:
            return umsgpack.packb(self.payload)
        except umsgpack.PackException as exc:
            raise MessageError('Could not pack msgpack payload') from exc

    @classmethod
    def _unpack_payload(cls, payload):
        try:
            return umsgpack.unpackb(payload)
        except (umsgpack.UnpackException, TypeError) as exc:
            raise MessageError('Could not unpack msgpack payload') from exc

    @classmethod
    def _encrypt_payload(cls, client, nonce, payload):
        try:
            _, data = client.box.encrypt(payload, nonce=nonce, pack_nonce=False)
            return data
        except ValueError as exc:
            raise MessageError('Could not encrypt payload') from exc

    @classmethod
    def _decrypt_payload(cls, client, nonce, data):
        try:
            return client.box.decrypt(data, nonce=nonce)
        except ValueError as exc:
            raise MessageError('Could not decrypt payload') from exc


class RawMessage(AbstractMessage):
    type = 'raw'  # Note: This field is used for logging purposes only

    def __init__(
            self, source, destination, data,
            source_type=None, destination_type=None
    ):
        super().__init__(
            source, destination,
            source_type=source_type, destination_type=destination_type
        )
        self._data = data

    def __str__(self):
        return _message_representation(
            self.__class__.__name__, self._nonce, self._data)

    def pack(self, client):
        return self._data

    def prepare_payload(self, client, nonce):
        return

    @classmethod
    def unpack(cls, client, data):
        return AbstractBaseMessage.unpack(client, data)

    @classmethod
    def check_payload(cls, client, payload):
        pass


class ServerHelloMessage(AbstractBaseMessage):
    type = MessageType.server_hello
    encrypted = False

    @classmethod
    def create(cls, source, destination, server_public_key):
        # noinspection PyCallingNonCallable
        return cls(source, destination, {
            'type': cls.type.value,
            'key': server_public_key,
        })

    @classmethod
    def check_payload(cls, client, payload):
        """
        MessageError
        """
        payload['key'] = payload.get('key')
        validate_public_key(payload['key'])
        return payload

    @property
    def server_public_key(self):
        return self.payload['key']


class ClientHelloMessage(AbstractBaseMessage):
    type = MessageType.client_hello
    encrypted = False

    @classmethod
    def create(cls, source, destination, client_public_key):
        # noinspection PyCallingNonCallable
        return cls(source, destination, {
            'type': cls.type.value,
            'key': client_public_key,
        })

    @classmethod
    def check_payload(cls, client, payload):
        """
        MessageError
        """
        payload['key'] = payload.get('key')
        validate_public_key(payload['key'])
        return payload

    @property
    def client_public_key(self):
        return self.payload['key']


class ClientAuthMessage(AbstractBaseMessage):
    type = MessageType.client_auth
    encrypted = True

    @classmethod
    def create(
            cls, source, destination, server_cookie, subprotocols,
            ping_interval=None, server_key=None):
        # noinspection PyCallingNonCallable
        payload = {
            'type': cls.type.value,
            'your_cookie': server_cookie,
            'subprotocols': subprotocols,
        }
        if ping_interval is not None:
            payload['ping_interval'] = ping_interval
        if server_key is not None:
            payload['your_key'] = server_key
        return cls(source, destination, payload)

    @classmethod
    def check_payload(cls, client, payload):
        """
        MessageError
        """
        validate_cookie(payload.get('your_cookie'))
        validate_subprotocols(payload.get('subprotocols'))
        ping_interval = payload.get('ping_interval')
        if ping_interval is not None:
            validate_ping_interval(ping_interval)
        server_key = payload.get('your_key')
        if server_key is not None:
            validate_public_key(server_key)
        return payload

    @property
    def server_cookie(self):
        return self.payload['your_cookie']

    @property
    def subprotocols(self):
        return self.payload['subprotocols']

    @property
    def ping_interval(self):
        return self.payload.get('ping_interval')

    @property
    def server_key(self):
        return self.payload.get('your_key')


class ServerAuthMessage(AbstractBaseMessage):
    type = MessageType.server_auth
    encrypted = True

    @classmethod
    def create(
            cls, source, destination, client_cookie,
            sign_keys=False, initiator_connected=None, responder_ids=None
    ):
        payload = {
            'type': cls.type.value,
            'your_cookie': client_cookie,
        }
        if initiator_connected is not None:
            payload['initiator_connected'] = initiator_connected
        if responder_ids is not None:
            payload['responders'] = responder_ids
        # noinspection PyCallingNonCallable
        return cls(source, destination, payload, extra={'sign_keys': sign_keys})

    def prepare_payload(self, client, nonce):
        """
        Late-signing of the keys.

        Raises :exc:`MessageError` in case the keys could not be
        signed.
        """
        if self.extra.get('sign_keys', False):
            self.payload['signed_keys'] = sign_keys_(client, nonce)

    @classmethod
    def check_payload(cls, client, payload):
        """
        MessageError
        """
        validate_cookie(payload.get('your_cookie'))
        signed_keys = payload.get('signed_keys')
        if signed_keys is not None:
            validate_signed_keys(signed_keys)
        responders = payload.get('responders')
        if responders is not None:
            validate_responder_ids(responders)
        initiator_connected = payload.get('initiator_connected')
        if initiator_connected is not None:
            validate_initiator_connected(initiator_connected)
        return payload

    @property
    def client_cookie(self):
        return self.payload['your_cookie']

    @property
    def signed_keys(self):
        """
        KeyError in case the message contains no 'signed_keys' field.
        """
        return self.payload['signed_keys']

    @property
    def initiator_connected(self):
        """
        KeyError in case the message is directed at an initiator.
        """
        return self.payload['initiator_connected']

    @property
    def responder_ids(self):
        """
        KeyError in case the message is directed at a responder.
        """
        return self.payload['responders']


class NewInitiatorMessage(AbstractBaseMessage):
    type = MessageType.new_initiator
    encrypted = True

    @classmethod
    def create(cls, source, destination):
        # noinspection PyCallingNonCallable
        return cls(source, destination, {
            'type': cls.type.value,
        })

    @classmethod
    def check_payload(cls, client, payload):
        return payload


class NewResponderMessage(AbstractBaseMessage):
    type = MessageType.new_responder
    encrypted = True

    @classmethod
    def create(cls, source, destination, responder_id):
        # noinspection PyCallingNonCallable
        return cls(source, destination, {
            'type': cls.type.value,
            'id': responder_id,
        })

    @classmethod
    def check_payload(cls, client, payload):
        """
        MessageError
        """
        validate_responder_id(payload.get('id'))
        return payload

    @property
    def responder_id(self):
        return self.payload['id']


class DropResponderMessage(AbstractBaseMessage):
    type = MessageType.drop_responder
    encrypted = True

    @classmethod
    def create(cls, source, destination, responder_id, reason=None):
        if reason is None:
            reason = CloseCode.drop_by_initiator
        # noinspection PyCallingNonCallable
        return cls(source, destination, {
            'type': cls.type.value,
            'id': responder_id,
            'reason': reason.value
        })

    @classmethod
    def check_payload(cls, client, payload):
        """
        MessageError
        """
        validate_responder_id(payload.get('id'))
        payload['reason'] = validate_drop_reason(payload.get('reason'))
        return payload

    @property
    def responder_id(self):
        return self.payload['id']

    @property
    def reason(self):
        return self.payload['reason']


class SendErrorMessage(AbstractBaseMessage):
    type = MessageType.send_error
    encrypted = True

    @classmethod
    def create(cls, source, destination, message_id):
        # noinspection PyCallingNonCallable
        return cls(source, destination, {
            'type': cls.type.value,
            'id': message_id,
        })

    @classmethod
    def check_payload(cls, client, payload):
        """
        MessageError
        """
        validate_hash(payload.get('hash'))
        return payload

    @property
    def message_hash(self):
        return self.payload['hash']



"""
This is a SaltyRTC server implementation for Python 3.4+ using
:mod:`asyncio`.
"""
import itertools

# noinspection PyUnresolvedReferences
from .exception import *  # noqa
# noinspection PyUnresolvedReferences
from .common import *  # noqa
# noinspection PyUnresolvedReferences
from .message import *  # noqa
# noinspection PyUnresolvedReferences
from .protocol import *  # noqa
# noinspection PyUnresolvedReferences
from .server import *  # noqa
# noinspection PyUnresolvedReferences
from .util import *  # noqa
# noinspection PyUnresolvedReferences
from .events import *  # noqa

__all__ = tuple(itertools.chain(
    ('bin',),
    exception.__all__,  # noqa
    common.__all__,  # noqa
    message.__all__,  # noqa
    protocol.__all__,  # noqa
    server.__all__,  # noqa
    util.__all__,  # noqa
    events.__all__,  # noqa
))

__author__ = 'Lennart Grahl <lennart.grahl@gmail.com>'
__status__ = 'Production'
__version__ = '1.0.0'



import asyncio
import binascii
import inspect

from collections import OrderedDict
from typing import (
    Dict,
    List,
)

import websockets

from . import util
from .common import (
    RELAY_TIMEOUT,
    AddressType,
    CloseCode,
    MessageType,
    SubProtocol,
)
from .events import (
    Event,
    EventRegistry,
)
from .exception import (
    Disconnected,
    DowngradeError,
    MessageError,
    MessageFlowError,
    PathError,
    PingTimeoutError,
    ServerKeyError,
    SignalingError,
    SlotsFullError,
)
from .message import (
    NewInitiatorMessage,
    NewResponderMessage,
    RawMessage,
    SendErrorMessage,
    ServerAuthMessage,
    ServerHelloMessage,
)
from .protocol import (
    Path,
    PathClient,
    Protocol,
)

try:
    from collections.abc import Coroutine
except ImportError:  # python 3.4
    # noinspection PyPackageRequirements
    from backports_abc import Coroutine

__all__ = (
    'serve',
    'ServerProtocol',
    'Paths',
    'Server',
)


@asyncio.coroutine
def serve(
        ssl_context, keys, paths=None, host=None, port=8765, loop=None,
        event_callbacks: Dict[Event, List[Coroutine]] = None, server_class=None
):
    """
    Start serving SaltyRTC Signalling Clients.

    Arguments:
        - `ssl_context`: An `ssl.SSLContext` instance for WSS.
        - `keys`: A sorted iterable of :class:`libnacl.public.SecretKey`
          instances containing permanent private keys of the server.
          The first key will be designated as the primary key.
        - `paths`: A :class:`Paths` instance that maps path names to
          :class:`Path` instances. Can be used to share paths on
          multiple WebSockets. Defaults to an empty paths instance.
        - `host`: The hostname or IP address the server will listen on.
          Defaults to all interfaces.
        - `port`: The port the client should connect to. Defaults to
          `8765`.
        - `loop`: A :class:`asyncio.BaseEventLoop` instance or `None`
          if the default event loop should be used.
        - `event_callbacks`: An optional dict with keys being an
          :class:`Event` and the value being a list of callback
          coroutines. The callback will be called every time the event
          occurs.
        - `server_class`: An optional :class:`Server` class to create
          an instance from.

    Raises :exc:`ServerKeyError` in case one or more keys have been repeated.
    """
    if loop is None:
        loop = asyncio.get_event_loop()

    # Create paths if not given
    if paths is None:
        paths = Paths()

    # Create server
    if server_class is None:
        server_class = Server
    server = server_class(keys, paths, loop=loop)

    # Register event callbacks
    if event_callbacks is not None:
        for event, callbacks in event_callbacks.items():
            for callback in callbacks:
                server.register_event_callback(event, callback)

    # Start server
    ws_server = yield from websockets.serve(
        server.handler,
        ssl=ssl_context,
        host=host,
        port=port,
        subprotocols=server.subprotocols
    )

    # Set server instance
    server.server = ws_server

    # Return server
    return server


class ServerProtocol(Protocol):
    __slots__ = (
        '_log',
        '_loop',
        '_server',
        'subprotocol',
        'path',
        'client',
        'handler_task'
    )

    def __init__(self, server, subprotocol, loop=None):
        self._log = util.get_logger('server.protocol')
        self._loop = asyncio.get_event_loop() if loop is None else loop

        # Server instance and subprotocol
        self._server = server
        self.subprotocol = subprotocol

        # Path and client instance
        self.path = None
        self.client = None

        # Handler task that is set after 'connection_made' has been called
        self.handler_task = None

        # Determine subprotocol selection function
        # Might be a static method, might be a normal method, see
        # https://github.com/aaugustin/websockets/pull/132
        protocol = websockets.WebSocketServerProtocol
        select_subprotocol = inspect.getattr_static(protocol, 'select_subprotocol')
        if isinstance(select_subprotocol, staticmethod):
            self._select_subprotocol = protocol.select_subprotocol
        else:
            def _select_subprotocol(client_subprotocols, server_subprotocols):
                # noinspection PyTypeChecker
                return protocol.select_subprotocol(
                    None, client_subprotocols, server_subprotocols)
            self._select_subprotocol = _select_subprotocol

    def connection_made(self, connection, ws_path):
        self.handler_task = self._loop.create_task(self.handler(connection, ws_path))

    @asyncio.coroutine
    def close(self, code=1000):
        # Note: The client will be set as early as possible without any yielding.
        #       Thus, self.client is either set and can be closed or the connection
        #       is already closing (see the corresponding lines in 'handler' and
        #       'get_path_client')
        if self.client is not None:
            yield from self.client.close(code=code)

    @asyncio.coroutine
    def handler(self, connection, ws_path):
        self._log.debug('New connection on WS path {}', ws_path)

        # Get path and client instance as early as possible
        try:
            path, client = self.get_path_client(connection, ws_path)
        except PathError as exc:
            self._log.notice('Closing due to path error: {}', exc)
            yield from connection.close(code=CloseCode.protocol_error.value)
            self._server.raise_event(
                Event.disconnected, None, CloseCode.protocol_error.value)
            return
        client.log.info('Connection established')
        client.log.debug('Worker started')

        # Store path and client
        self.path = path
        self.client = client
        self._server.register(self)

        # Handle client until disconnected or an exception occurred
        hex_path = binascii.hexlify(self.path.initiator_key).decode('ascii')
        try:
            yield from self.handle_client()
        except Disconnected as exc:
            client.log.info('Connection closed')
            self._server.raise_event(Event.disconnected, hex_path, exc.reason)
        except SlotsFullError as exc:
            client.log.notice('Closing because all path slots are full: {}', exc)
            yield from client.close(code=CloseCode.path_full_error.value)
            self._server.raise_event(
                    Event.disconnected, hex_path, CloseCode.path_full_error.value)
        except ServerKeyError as exc:
            client.log.notice('Closing due to server key error: {}', exc)
            yield from client.close(code=CloseCode.invalid_key.value)
            self._server.raise_event(
                    Event.disconnected, hex_path, CloseCode.invalid_key.value)
        except SignalingError as exc:
            client.log.notice('Closing due to protocol error: {}', exc)
            yield from client.close(code=CloseCode.protocol_error.value)
            self._server.raise_event(
                    Event.disconnected, hex_path, CloseCode.protocol_error.value)
        except Exception as exc:
            client.log.exception('Closing due to exception:', exc)
            yield from client.close(code=CloseCode.internal_error.value)
            self._server.raise_event(
                    Event.disconnected, hex_path, CloseCode.internal_error.value)
        else:
            client.log.error('Client closed without exception')

        # Remove client from path
        path.remove_client(client)

        # Remove protocol from server and stop
        self._server.unregister(self)
        client.log.debug('Worker stopped')

    def get_path_client(self, connection, ws_path):
        # Extract public key from path
        initiator_key = ws_path[1:]

        # Validate key
        if len(initiator_key) != self.PATH_LENGTH:
            raise PathError('Invalid path length: {}'.format(len(initiator_key)))
        try:
            initiator_key = binascii.unhexlify(initiator_key)
        except (binascii.Error, ValueError) as exc:
            raise PathError('Could not unhexlify path') from exc

        # Get path instance
        path = self._server.paths.get(initiator_key)

        # Create client instance
        client = PathClient(connection, path.number, initiator_key,
                            loop=self._loop)

        # Return path and client
        return path, client

    @asyncio.coroutine
    def handle_client(self):
        """
        SignalingError
        PathError
        Disconnected
        MessageError
        MessageFlowError
        SlotsFullError
        DowngradeError
        ServerKeyError
        """
        client = self.client

        # Do handshake
        client.log.debug('Starting handshake')
        yield from self.handshake()
        client.log.info('Handshake completed')

        # Task: Execute enqueued tasks
        client.log.debug('Starting poll for enqueued tasks task')
        task_loop_task = self._loop.create_task(self.task_loop())
        tasks = [task_loop_task]
        coroutines = []

        # Task: Poll for messages
        hex_path = binascii.hexlify(self.path.initiator_key).decode('ascii')
        if client.type == AddressType.initiator:
            client.log.debug('Starting runner for initiator')
            self._server.raise_event(Event.initiator_connected, hex_path)
            coroutines.append(self.initiator_receive_loop())
        elif client.type == AddressType.responder:
            client.log.debug('Starting runner for responder')
            self._server.raise_event(Event.responder_connected, hex_path)
            coroutines.append(self.responder_receive_loop())
        else:
            raise ValueError('Invalid address type: {}'.format(client.type))

        # Task: Keep alive
        client.log.debug('Starting keep-alive task')
        coroutines.append(self.keep_alive_loop())

        # Wait until complete
        tasks += [self._loop.create_task(coroutine) for coroutine in coroutines]
        while True:
            done, pending = yield from asyncio.wait(
                tasks, loop=self._loop, return_when=asyncio.FIRST_COMPLETED)
            for task in done:
                client.log.debug('Task done {}', done)
                exc = task.exception()

                # Cancel pending tasks
                for pending_task in pending:
                    client.log.debug('Cancelling task {}', pending_task)
                    pending_task.cancel()

                # Raise (or re-raise)
                if exc is None:
                    if task == task_loop_task:
                        # Task loop may return early and it's okay
                        client.log.debug('Task loop returned early')
                        tasks.remove(task_loop_task)
                    else:
                        client.log.error('Task {} returned unexpectedly', task)
                        raise SignalingError('A task returned unexpectedly')
                else:
                    raise exc

    @asyncio.coroutine
    def handshake(self):
        """
        Disconnected
        MessageError
        MessageFlowError
        SlotsFullError
        DowngradeError
        ServerKeyError
        """
        client = self.client

        # Send server-hello
        message = ServerHelloMessage.create(
            AddressType.server, client.id, client.server_key.pk)
        client.log.debug('Sending server-hello')
        yield from client.send(message)

        # Receive client-hello or client-auth
        client.log.debug('Waiting for client-hello')
        message = yield from client.receive()
        if message.type == MessageType.client_auth:
            client.log.debug('Received client-auth')
            # Client is the initiator
            client.type = AddressType.initiator
            yield from self.handshake_initiator(message)
        elif message.type == MessageType.client_hello:
            client.log.debug('Received client-hello')
            # Client is a responder
            client.type = AddressType.responder
            yield from self.handshake_responder(message)
        else:
            error = "Expected 'client-hello' or 'client-auth', got '{}'"
            raise MessageFlowError(error.format(message.type))

    @asyncio.coroutine
    def handshake_initiator(self, message):
        """
        Disconnected
        MessageError
        MessageFlowError
        DowngradeError
        ServerKeyError
        """
        path, initiator = self.path, self.client

        # Handle client-auth
        self._handle_client_auth(message)

        # Authenticated
        previous_initiator = path.set_initiator(initiator)
        if previous_initiator is not None:
            # Drop previous initiator using the task queue of the previous initiator
            path.log.debug('Dropping previous initiator {}', previous_initiator)
            previous_initiator.log.debug('Dropping (another initiator connected)')
            coroutine = previous_initiator.close(code=CloseCode.drop_by_initiator.value)
            yield from previous_initiator.enqueue_task(coroutine)

        # Send new-initiator message if any responder is present
        responder_ids = path.get_responder_ids()
        for responder_id in responder_ids:
            responder = path.get_responder(responder_id)

            # Create message and add send coroutine to task queue of the responder
            message = NewInitiatorMessage.create(AddressType.server, responder_id)
            responder.log.debug('Enqueueing new-initiator message')
            yield from responder.enqueue_task(responder.send(message))

        # Send server-auth
        responder_ids = path.get_responder_ids()
        message = ServerAuthMessage.create(
            AddressType.server, initiator.id, initiator.cookie_in,
            sign_keys=len(self._server.keys) > 0, responder_ids=responder_ids)
        initiator.log.debug('Sending server-auth including responder ids')
        yield from initiator.send(message)

    @asyncio.coroutine
    def handshake_responder(self, message):
        """
        Disconnected
        MessageError
        MessageFlowError
        SlotsFullError
        DowngradeError
        ServerKeyError
        """
        path, responder = self.path, self.client

        # Set key on client
        responder.set_client_key(message.client_public_key)

        # Receive client-auth
        message = yield from responder.receive()
        if message.type != MessageType.client_auth:
            error = "Expected 'client-auth', got '{}'"
            raise MessageFlowError(error.format(message.type))

        # Handle client-auth
        self._handle_client_auth(message)

        # Authenticated
        id_ = path.add_responder(responder)

        # Send new-responder message if initiator is present
        initiator = path.get_initiator()
        initiator_connected = initiator is not None
        if initiator_connected:
            # Create message and add send coroutine to task queue of the initiator
            message = NewResponderMessage.create(AddressType.server, initiator.id, id_)
            initiator.log.debug('Enqueueing new-responder message')
            yield from initiator.enqueue_task(initiator.send(message))

        # Send server-auth
        message = ServerAuthMessage.create(
            AddressType.server, responder.id, responder.cookie_in,
            sign_keys=len(self._server.keys) > 0,
            initiator_connected=initiator_connected)
        responder.log.debug('Sending server-auth without responder ids')
        yield from responder.send(message)

    @asyncio.coroutine
    def task_loop(self):
        client = self.client
        while not client.connection_closed.done():
            # Get a task from the queue
            task = yield from client.dequeue_task()

            # Wait and catch exceptions, ignore cancelled tasks
            client.log.debug('Waiting for task to complete {}', task)
            try:
                yield from task
            except asyncio.CancelledError:
                client.log.debug('Task cancelled {}', task)

    @asyncio.coroutine
    def initiator_receive_loop(self):
        path, initiator = self.path, self.client
        while not initiator.connection_closed.done():
            # Receive relay message or drop-responder
            message = yield from initiator.receive()

            # Relay
            if isinstance(message, RawMessage):
                # Lookup responder
                responder = path.get_responder(message.destination)
                # Send to responder
                yield from self.relay_message(responder, message.destination, message)
            # Drop-responder
            elif message.type == MessageType.drop_responder:
                # Lookup responder
                responder = path.get_responder(message.responder_id)
                if responder is not None:
                    # Drop responder using its task queue
                    path.log.debug(
                        'Dropping responder {}, reason: {}', responder, message.reason)
                    responder.log.debug(
                        'Dropping (requested by initiator), reason: {}', message.reason)
                    coroutine = responder.close(code=message.reason.value)
                    yield from responder.enqueue_task(coroutine)
                else:
                    log_message = 'Responder {} already dropped, nothing to do'
                    path.log.debug(log_message, responder)
            else:
                error = "Expected relay message or 'drop-responder', got '{}'"
                raise MessageFlowError(error.format(message.type))

    @asyncio.coroutine
    def responder_receive_loop(self):
        path, responder = self.path, self.client
        while not responder.connection_closed.done():
            # Receive relay message
            message = yield from responder.receive()

            # Relay
            if isinstance(message, RawMessage):
                # Lookup initiator
                initiator = path.get_initiator()
                # Send to initiator
                yield from self.relay_message(initiator, AddressType.initiator, message)
            else:
                error = "Expected relay message, got '{}'"
                raise MessageFlowError(error.format(message.type))

    @asyncio.coroutine
    def relay_message(self, destination, destination_id, message):
        source = self.client

        # Prepare message
        source.log.debug('Packing relay message')
        message_id = message.pack(source)[16:]

        @asyncio.coroutine
        def send_error_message():
            # Create message and add send coroutine to task queue of the source
            error = SendErrorMessage.create(
                AddressType.server, source.id, message_id)
            source.log.info('Relaying failed, enqueuing send-error')
            yield from source.enqueue_task(source.send(error))

        # Destination not connected? Send 'send-error' to source
        if destination is None:
            error_message = ('Cannot relay message, no connection for '
                             'destination id 0x{:02x}')
            source.log.info(error_message, destination_id)
            yield from send_error_message()
            return

        # Add send task to task queue of the source
        task = self._loop.create_task(destination.send(message))
        destination.log.debug('Enqueueing relayed message from 0x{:02x}', source.id)
        yield from destination.enqueue_task(task)

        # noinspection PyBroadException
        try:
            # Wait for send task to complete
            yield from asyncio.wait_for(task, RELAY_TIMEOUT, loop=self._loop)
        except asyncio.TimeoutError:
            # Timed out, send 'send-error' to source
            log_message = 'Sending relayed message to 0x{:02x} timed out'
            source.log.info(log_message, destination.id)
            yield from send_error_message()
        except Exception:
            # An exception has been triggered while sending the message.
            # Note: We don't care about the actual exception as the task
            #       will also trigger that exception on the destination
            #       client's handler who will log what happened.
            log_message = 'Sending relayed message failed, receiver 0x{:02x} is gone'
            source.log.info(log_message, destination.id)
            yield from send_error_message()

    @asyncio.coroutine
    def keep_alive_loop(self):
        """
        Disconnected
        PingTimeoutError
        """
        client = self.client
        while not client.connection_closed.done():
            # Wait
            yield from asyncio.sleep(client.keep_alive_interval, loop=self._loop)

            # Send ping and wait for pong
            client.log.debug('Ping')
            try:
                pong_future = yield from client.ping()
                yield from asyncio.wait_for(
                    pong_future, client.keep_alive_timeout, loop=self._loop)
            except asyncio.TimeoutError:
                raise PingTimeoutError(client)
            else:
                client.log.debug('Pong')
                client.keep_alive_pings += 1

    def _handle_client_auth(self, message):
        """
        MessageError
        DowngradeError
        ServerKeyError
        """
        client = self.client

        # Validate cookie and ensure no sub-protocol downgrade took place
        self._validate_cookie(message.server_cookie, client.cookie_out)
        self._validate_subprotocol(message.subprotocols)

        # Set the keep alive interval (if any)
        if message.ping_interval is not None:
            client.log.debug('Setting keep-alive interval to {}', message.ping_interval)
            client.keep_alive_interval = message.ping_interval

        # Set the public permanent key the client wants to use (or fallback to primary)
        server_keys_count = len(self._server.keys)
        if message.server_key is not None:
            # No permanent key pair?
            if server_keys_count == 0:
                raise ServerKeyError('Server does not have a permanent public key')

            # Find the key instance
            server_key = self._server.keys.get(message.server_key)
            if server_key is None:
                raise ServerKeyError(
                    'Server does not have the requested permanent public key')

            # Set the key instance on the client
            client.server_permanent_key = server_key
        elif server_keys_count > 0:
            # Use primary permanent key
            client.server_permanent_key = next(iter(self._server.keys.values()))

    def _validate_cookie(self, expected_cookie, actual_cookie):
        """
        MessageError
        """
        self.client.log.debug('Validating cookie')
        if not util.consteq(expected_cookie, actual_cookie):
            raise MessageError('Cookies do not match')

    def _validate_subprotocol(self, client_subprotocols):
        """
        MessageError
        DowngradeError
        """
        self.client.log.debug(
            'Checking for subprotocol downgrade, client: {}, server: {}',
            client_subprotocols, self._server.subprotocols)
        chosen = self._select_subprotocol(
            client_subprotocols, self._server.subprotocols)
        if chosen != self.subprotocol.value:
            raise DowngradeError('Subprotocol downgrade detected')


class Paths:
    __slots__ = ('_log', 'number', 'paths')

    def __init__(self):
        self._log = util.get_logger('paths')
        self.number = 0
        self.paths = {}

    def get(self, initiator_key):
        if self.paths.get(initiator_key) is None:
            self.number += 1
            self.paths[initiator_key] = Path(initiator_key, self.number)
            self._log.debug('Created new path: {}', self.number)
        return self.paths[initiator_key]

    def clean(self, path):
        if path.empty:
            try:
                del self.paths[path.initiator_key]
            except KeyError:
                self._log.warning('Path {} has already been removed', path.number)
            else:
                self._log.debug('Removed empty path: {}', path.number)


class Server(asyncio.AbstractServer):
    subprotocols = [
        SubProtocol.saltyrtc_v1.value
    ]

    def __init__(self, keys, paths, loop=None):
        self._log = util.get_logger('server')
        self._loop = asyncio.get_event_loop() if loop is None else loop

        # WebSocket server instance
        self._server = None

        # Validate & store keys
        if keys is None:
            keys = []
        if len(keys) != len({key.pk for key in keys}):
            raise ServerKeyError('Repeated permanent keys')
        self.keys = OrderedDict(((key.pk, key) for key in keys))

        # Store paths
        self.paths = paths

        # Store server protocols
        self.protocols = set()

        # Event Registry
        self._events = EventRegistry()

    @property
    def server(self):
        return self._server

    @server.setter
    def server(self, server):
        self._server = server
        self._log.debug('Server instance: {}', server)

    @asyncio.coroutine
    def handler(self, connection, ws_path):
        # Convert sub-protocol
        try:
            subprotocol = SubProtocol(connection.subprotocol)
        except ValueError:
            subprotocol = None

        # Determine ServerProtocol instance by selected sub-protocol
        if subprotocol != SubProtocol.saltyrtc_v1:
            self._log.notice('Could not negotiate a sub-protocol, dropping client')
            # We need to close the connection manually as the client may choose
            # to ignore
            yield from connection.close(code=CloseCode.subprotocol_error.value)
            self.raise_event(Event.disconnected, None, CloseCode.subprotocol_error.value)
        else:
            protocol = ServerProtocol(self, subprotocol, loop=self._loop)
            protocol.connection_made(connection, ws_path)
            yield from protocol.handler_task

    def register(self, protocol):
        self.protocols.add(protocol)
        self._log.debug('Protocol registered: {}', protocol)

    def unregister(self, protocol):
        self.protocols.remove(protocol)
        self._log.debug('Protocol unregistered: {}', protocol)
        self.paths.clean(protocol.path)

    def register_event_callback(self, event: Event, callback: Coroutine):
        """
        Register a new event callback.
        """
        self._events.register(event, callback)

    def raise_event(self, event: Event, *data):
        """
        Raise an event and call all registered event callbacks.
        """
        for callback in self._events.get_callbacks(event):
            self._loop.create_task(callback(event, *data))

    def close(self):
        """
        Close open connections and the server.
        """
        self._loop.create_task(self._close_after_all_protocols_closed())

    @asyncio.coroutine
    def wait_closed(self):
        """
        Wait until all connections and the server itself is closed.
        """
        yield from self._wait_connections_closed()
        yield from self.server.wait_closed()

    @asyncio.coroutine
    def _wait_connections_closed(self):
        """
        Wait until all connections to the server have been closed.
        """
        if len(self.protocols) > 0:
            tasks = [protocol.handler_task for protocol in self.protocols]
            yield from asyncio.wait(tasks, loop=self._loop)

    @asyncio.coroutine
    def _close_after_all_protocols_closed(self, timeout=None):
        # Schedule closing all protocols
        self._log.debug('Closing protocols')
        if len(self.protocols) > 0:
            tasks = [protocol.close(code=CloseCode.going_away.value)
                     for protocol in self.protocols]

            # Wait until all protocols are closed (we need the server to be active for the
            # WebSocket close protocol)
            yield from asyncio.wait(tasks, loop=self._loop, timeout=timeout)

        # Now we can close the server
        self._log.debug('Closing server')
        self.server.close()



"""
Contains all exceptions used for the SaltyRTC server.
"""
__all__ = (
    'SignalingError',
    'InternalError',
    'PathError',
    'SlotsFullError',
    'ServerKeyError',
    'MessageFlowError',
    'PingTimeoutError',
    'Disconnected',
    'MessageError',
    'DowngradeError',
)


class SignalingError(Exception):
    """
    General error of this module. All other exceptions are derived from
    this class.
    """


class InternalError(SignalingError):
    """
    Server misbehaved. Always report occurrences of this error!
    """


class PathError(SignalingError):
    """
    TODO: Describe
    """


class SlotsFullError(SignalingError):
    """
    No free slot for a responder.
    """


class ServerKeyError(SignalingError):
    """
    Raised when the server either does not have a permanent key pair
    but the client requested one or the server does not have the key
    pair that has been requested.
    """


class MessageFlowError(SignalingError):
    """
    TODO: Describe
    """


class PingTimeoutError(SignalingError):
    """
    TODO: Describe
    """
    def __init__(self, client):
        self.client = client

    def __str__(self):
        return 'Ping to {} timed out'.format(*self.args)


class Disconnected(Exception):
    """
    TODO: Describe
    """
    def __init__(self, reason: int = None):
        self.reason = reason


class MessageError(SignalingError):
    """
    Raised when a message is invalid.
    """


class DowngradeError(SignalingError):
    """
    A protocol downgrade has been detected.
    """



import asyncio
import os
import signal
import sys

import logbook.more

import saltyrtc.server


def env(name, default=None):
    return os.environ.get(name, default)


def require_env(name):
    value = env(name)
    if value is None:
        print("Missing '{}' env variable".format(name))
        sys.exit(1)
    return value


def main():
    """
    Run the SaltyRTC server until Ctrl+C has been pressed.

    The signal *HUP* will restart the server.
    """
    # Get event loop
    loop = asyncio.get_event_loop()

    while True:
        # Create SSL context
        if env('SALTYRTC_DISABLE_TLS') != 'yes-and-i-know-what-im-doing':
            ssl_context = saltyrtc.server.create_ssl_context(
                certfile=require_env('SALTYRTC_TLS_CERT'),
                keyfile=require_env('SALTYRTC_TLS_KEY'),
                dh_params_file=require_env('SALTYRTC_DH_PARAMS'),
            )
        else:
            ssl_context = None

        # Get permanent key
        if env('SALTYRTC_DISABLE_SERVER_PERMANENT_KEY') != 'yes-and-i-know-what-im-doing':
            permanent_keys = [saltyrtc.server.load_permanent_key(
                require_env('SALTYRTC_SERVER_PERMANENT_KEY'))]
        else:
            permanent_keys = None

        # Start server
        port = int(env('SALTYRTC_PORT', '8765'))
        coroutine = saltyrtc.server.serve(ssl_context, permanent_keys, port=port)
        server = loop.run_until_complete(coroutine)

        # Restart server on HUP signal
        restart_signal = asyncio.Future(loop=loop)

        def _restart_signal_handler(*_):
            restart_signal.set_result(True)

        # Register restart server routine
        loop.add_signal_handler(signal.SIGHUP, _restart_signal_handler)

        # Wait until Ctrl+C has been pressed
        try:
            loop.run_until_complete(restart_signal)
        except KeyboardInterrupt:
            pass

        # Remove the signal handler
        loop.remove_signal_handler(signal.SIGHUP)

        # Wait until server is closed and close the event loop
        server.close()
        loop.run_until_complete(server.wait_closed())

        # Stop?
        if not restart_signal.done():
            break

    # Close loop
    loop.close()


if __name__ == '__main__':
    # Enable asyncio debug logging
    os.environ['PYTHONASYNCIODEBUG'] = '1'

    # Enable logging
    saltyrtc.server.enable_logging(level=logbook.TRACE, redirect_loggers={
        'asyncio': logbook.DEBUG,
        'websockets': logbook.DEBUG,
    })

    # Run 'main'
    logging_handler = logbook.more.ColorizedStderrHandler()
    with logging_handler.applicationbound():
        main()



"""
This package will be moved to `saltyrtc.common` as soon as a SaltyRTC
client is being written.
"""
import enum

from .exception import MessageError

__all__ = (
    'DATA_LENGTH_MIN',
    'KEY_LENGTH',
    'NONCE_LENGTH',
    'NONCE_FORMATTER',
    'COOKIE_LENGTH',
    'HASH_LENGTH',
    'SIGNED_KEYS_CIPHERTEXT_LENGTH',
    'RELAY_TIMEOUT',
    'KEEP_ALIVE_INTERVAL_MIN',
    'KEEP_ALIVE_INTERVAL_DEFAULT',
    'KEEP_ALIVE_TIMEOUT',
    'OverflowSentinel',
    'SubProtocol',
    'CloseCode',
    'AddressType',
    'MessageType',
    'available_slot_range',
    'is_initiator_id',
    'is_responder_id',
    'validate_public_key',
    'validate_subprotocols',
    'validate_cookie',
    'validate_signed_keys',
    'validate_initiator_connected',
    'validate_responder_id',
    'validate_responder_ids',
    'validate_hash',
    'validate_ping_interval',
    'validate_drop_reason',
    'sign_keys',
)

DATA_LENGTH_MIN = 25
KEY_LENGTH = 32
NONCE_LENGTH = 24
NONCE_FORMATTER = '!16s2B6s'
COOKIE_LENGTH = 16
HASH_LENGTH = 32
SIGNED_KEYS_CIPHERTEXT_LENGTH = 80
RELAY_TIMEOUT = 30.0
KEEP_ALIVE_INTERVAL_MIN = 1.0
KEEP_ALIVE_INTERVAL_DEFAULT = 3600.0
KEEP_ALIVE_TIMEOUT = 30.0


class OverflowSentinel:
    """
    The combined sequence number will be set to this object if the
    counter did overflow.
    """


@enum.unique
class SubProtocol(enum.Enum):
    saltyrtc_v1 = 'v1.saltyrtc.org'


# Valid drop responder reasons
_drop_reasons = {3001, 3002, 3004, 3005}


@enum.unique
class CloseCode(enum.IntEnum):
    going_away = 1001
    subprotocol_error = 1002
    path_full_error = 3000
    protocol_error = 3001
    internal_error = 3002
    handover = 3003
    drop_by_initiator = 3004
    initiator_could_not_decrypt = 3005
    no_shared_tasks = 3006
    invalid_key = 3007

    @property
    def is_valid_drop_reason(self):
        return self.value in _drop_reasons


@enum.unique
class AddressType(enum.IntEnum):
    server = 0x00
    initiator = 0x01
    responder = 0xff

    @classmethod
    def from_address(cls, address):
        if address > 0x01:
            return cls.responder
        else:
            return cls(address)


@enum.unique
class MessageType(enum.Enum):
    """left out client-to-client message types"""
    server_hello = 'server-hello'
    client_hello = 'client-hello'
    client_auth = 'client-auth'
    server_auth = 'server-auth'
    new_responder = 'new-responder'
    new_initiator = 'new-initiator'
    drop_responder = 'drop-responder'
    send_error = 'send-error'


def available_slot_range():
    return range(0x01, 0xff + 1)


def is_initiator_id(id_):
    return id_ == 0x01


def is_responder_id(id_):
    return 0x01 < id_ <= 0xff


def validate_public_key(key):
    if not isinstance(key, bytes) or len(key) != KEY_LENGTH:
        raise MessageError('Invalid key')


def validate_cookie(cookie):
    if not isinstance(cookie, bytes):
        raise MessageError('Invalid cookie: Must be `bytes` instance')
    if len(cookie) != COOKIE_LENGTH:
        raise MessageError('Invalid cookie: Invalid length ({} != {})'.format(
            len(cookie), COOKIE_LENGTH))


def validate_subprotocols(subprotocols):
    try:
        iter(subprotocols)
    except TypeError as exc:
        raise MessageError('Sub-protocol list is not iterable') from exc


def validate_signed_keys(signed_keys):
    max_length = SIGNED_KEYS_CIPHERTEXT_LENGTH
    if not isinstance(signed_keys, bytes) or len(signed_keys) != max_length:
        raise MessageError("Invalid value for field 'signed_keys'")


def validate_initiator_connected(initiator_connected):
    if not isinstance(initiator_connected, bool):
        raise MessageError("Invalid value for field 'initiator_connected'")


def validate_responder_id(id_):
    if not is_responder_id(id_):
        raise MessageError('Invalid responder id in responder list')


def validate_responder_ids(ids):
    try:
        iterator = iter(ids)
    except TypeError as exc:
        raise MessageError('Responder list is not iterable') from exc
    for responder in iterator:
        validate_responder_id(responder)


def validate_hash(hash_):
    if not isinstance(hash_, bytes) or len(hash_) != HASH_LENGTH:
        raise MessageError('Invalid hash')


def validate_ping_interval(ping_interval):
    if not isinstance(ping_interval, int) or ping_interval < 0:
        raise MessageError('Invalid ping interval')


def validate_drop_reason(reason):
    # Default drop reason
    if reason is None:
        return CloseCode.drop_by_initiator

    # Validate reason
    try:
        reason = CloseCode(reason)
    except ValueError:
        raise MessageError('Invalid close code')
    if not reason.is_valid_drop_reason:
        raise MessageError('Reason not from acceptable range of close codes')

    return reason


def sign_keys(client, nonce):
    # Sign server's public session key and client's public permanent key (in that
    # order)
    payload = b''.join((client.server_key.pk, client.client_key))
    try:
        _, signed_keys = client.sign_box.encrypt(payload, nonce=nonce, pack_nonce=False)
        return signed_keys
    except ValueError as exc:
        raise MessageError('Could not sign keys') from exc



"""
The command line interface for the SaltyRTC signalling server.
"""
import asyncio
import enum
import os
import signal
import stat

import click
import libnacl.public

from . import __version__ as _version
from . import (
    server,
    util,
)

__all__ = (
    'cli',
    'version',
    'generate',
    'serve',
    'main',
)


def _h(text):
    """
    For some reason, :mod:`click` does not strip new line characters
    from helps in :func:`click.option` (although it does strip them
    from helps for :func:`click.command`). So, we have to do it
    ourselves.
    """
    return text.replace('\n', ' ')


def _get_logging_level(verbosity):
    # noinspection PyPackageRequirements
    import logbook
    return {
        1: logbook.CRITICAL,
        2: logbook.ERROR,
        3: logbook.WARNING,
        4: logbook.NOTICE,
        5: logbook.INFO,
        6: logbook.DEBUG,
        7: logbook.TRACE,
    }[verbosity]


class _ErrorCode(enum.IntEnum):
    safety_error = 2
    import_error = 3
    repeated_keys = 4


_logging_levels = 7


@click.group()
@click.option('-v', '--verbosity', type=click.IntRange(0, _logging_levels),
              default=0, help="Logging verbosity.")
@click.option('-c', '--colored', is_flag=True, help='Colourise logging output.')
@click.pass_context
def cli(ctx, verbosity, colored):
    """
    Command Line Interface. Use --help for details.
    """
    if verbosity > 0:
        try:
            # noinspection PyPackageRequirements,PyUnresolvedReferences
            import logbook
            # noinspection PyUnresolvedReferences,PyPackageRequirements
            import logbook.more
        except ImportError:
            click.echo('Please install saltyrtc.server[logging] for logging support.',
                       err=True)
            ctx.exit(code=_ErrorCode.import_error)

        # Translate logging level
        level = _get_logging_level(verbosity)

        # Enable asyncio debug logging if verbosity is high enough
        # noinspection PyUnboundLocalVariable
        if level <= logbook.DEBUG:
            os.environ['PYTHONASYNCIODEBUG'] = '1'

        # Enable logging
        util.enable_logging(level=level, redirect_loggers={
            'asyncio': level,
            'websockets': level,
        })

        # Get handler class
        if colored:
            handler_class = logbook.more.ColorizedStderrHandler
        else:
            handler_class = logbook.StderrHandler

        # Set up logging handler
        handler = handler_class(level=level)
        handler.push_application()
        ctx.obj['logging_handler'] = handler


@cli.command(short_help='Show version information.', help="""
Show the current version of the SaltyRTC signalling server and the
implemented protocol versions.
""")
def version():
    click.echo('Version: {}'.format(_version))
    click.echo('Protocols: {}'.format(server.Server.subprotocols))


@cli.command(short_help='Generate a new server permanent key pair.', help="""
Generate a new permanent key pair for the server and write the private key to
the respective KEY_FILE.
""")
@click.argument('KEY_FILE', type=click.Path(writable=True, dir_okay=False))
def generate(key_file):
    # Generate key pair
    key_pair = libnacl.public.SecretKey()

    # Write hex-encoded private key to file using proper permissions (0o400)
    perm_other = stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH
    perm_group = stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP
    current_umask = os.umask(perm_other | perm_group)
    with open(key_file, 'wb') as file:
        file.write(key_pair.hex_sk())
    os.umask(current_umask)

    # Print public key
    click.echo('Public permanent key: {}'.format(key_pair.hex_pk().decode('ascii')))


@cli.command(short_help='Start the signalling server.', help="""
Start the SaltyRTC signalling server. A HUP signal will restart the
server and reload the SSL certificate, the SSL private key and the
private permanent key of the server.""")
@click.option('-sc', '--sslcert', type=click.Path(exists=True), help=_h("""
Path to a PEM file that contains the SSL certificate."""))
@click.option('-sk', '--sslkey', type=click.Path(exists=True), help=_h("""
Path to a PEM file that contains the SSL private key. Will be read from
the SSL certificate file if not present."""))
@click.option('-dhp', '--dhparams', type=click.Path(exists=True), help=_h("""
Path to a PEM file that contains Diffie-Hellman parameters. Required
for DH(E) and ECDH(E) support."""))
@click.option('-k', '--key', type=click.Path(), multiple=True, help=_h("""
Path to a or a hex-encoded private permanent key of the server (e.g.
a NaCl private key). You can provide more than one key. The first key
provided will be used as the primary key."""))
@click.option('-h', '--host', help='Bind to a specific host.')
@click.option('-p', '--port', default=443, help='Listen on a specific port.')
@click.option('-l', '--loop', type=click.Choice(['asyncio', 'uvloop']), default='asyncio',
              help="Use a specific asyncio-compatible event loop. Defaults to 'asyncio'.")
@click.pass_context
def serve(ctx, **arguments):
    # Get arguments
    ssl_cert = arguments.get('sslcert', None)
    ssl_key = arguments.get('sslkey', None)
    dh_params = arguments.get('dhparams', None)
    keys = arguments.get('key')
    host = arguments.get('host')
    port = arguments['port']
    loop = arguments['loop']
    safety_off = os.environ.get('SALTYRTC_SAFETY_OFF') == 'yes-and-i-know-what-im-doing'

    # Make sure the user provides cert & keys or has safety turned off
    if ssl_cert is None or len(keys) == 0:
        if safety_off:
            click.echo(('It is RECOMMENDED to use SaltyRTC with both a SSL '
                       'certificate and a server permanent key!'), err=True)
        else:
            click.echo(('It is REQUIRED to provide a SSL certificate and a server '
                        'permanent key unless the environment variable '
                        "'SALTYRTC_SAFETY_OFF' is set to "
                        "'yes-and-i-know-what-im-doing'"), err=True)
            ctx.exit(code=_ErrorCode.safety_error)

    # Create SSL context
    ssl_context = None
    if ssl_cert is not None:
        ssl_context = util.create_ssl_context(
            certfile=ssl_cert, keyfile=ssl_key, dh_params_file=dh_params)

    # Get private permanent keys of the server
    if len(keys) > 0:
        keys = [util.load_permanent_key(key) for key in keys]

    # Validate permanent keys
    # Note: The permanent keys will be checked in the server coroutine but we don't want
    #       to look stupid.
    if len(keys) != len({key.pk for key in keys}):
        click.echo('At least one permanent key has been supplied more than once',
                   err=True)
        ctx.exit(code=_ErrorCode.repeated_keys)

    # Set event loop policy
    if loop == 'uvloop':
        try:
            # noinspection PyPackageRequirements,PyUnresolvedReferences
            import uvloop
        except ImportError:
            click.echo("Cannot use event loop 'uvloop', make sure it is installed.",
                       err=True)
            ctx.exit(code=_ErrorCode.import_error)
        # noinspection PyUnboundLocalVariable
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    # Get event loop
    loop = asyncio.get_event_loop()

    while True:
        # Run the server
        click.echo('Starting')
        if len(keys) > 0:
            primary_key, *secondary_keys = keys
            click.echo('Primary public permanent key: {}'.format(
                primary_key.hex_pk().decode('ascii')))
            for i, key in enumerate(secondary_keys, start=1):
                click.echo('Secondary key #{}: {}'.format(
                    i, key.hex_pk().decode('ascii')))
        coroutine = server.serve(ssl_context, keys, host=host, port=port, loop=loop)
        server_ = loop.run_until_complete(coroutine)

        # Restart server on HUP signal
        restart_signal = asyncio.Future(loop=loop)

        def _restart_signal_handler(*_):
            restart_signal.set_result(True)

        # Register restart server routine
        try:
            loop.add_signal_handler(signal.SIGHUP, _restart_signal_handler)
        except RuntimeError:
            click.echo('Cannot restart on SIGHUP, signal handler could not be added.')

        # Wait until Ctrl+C has been pressed
        click.echo('Started')
        try:
            loop.run_until_complete(restart_signal)
        except KeyboardInterrupt:
            click.echo()

        # Remove the signal handler
        loop.remove_signal_handler(signal.SIGHUP)

        # Close the server
        click.echo('Stopping')
        server_.close()
        loop.run_until_complete(server_.wait_closed())
        click.echo('Stopped')

        # Stop?
        if not restart_signal.done():
            restart_signal.cancel()
            break

    # Close loop
    loop.close()


def main():
    obj = {'logging_handler': None}
    try:
        cli(obj=obj)
    except Exception as exc:
        click.echo('An error occurred:', err=True)
        click.echo(exc, err=True)
        raise
    finally:
        if obj['logging_handler'] is not None:
            obj['logging_handler'].pop_application()



"""
TODO: Describe tests
"""
import libnacl.public
import pytest
import websockets


class TestPrerequisities:
    @pytest.mark.asyncio
    def test_server_handshake(self, ws_client_factory):
        """
        Make sure the server is reachable and we can do a simple
        WebSocket handshake using the correct sub-protocol.
        """
        client = yield from ws_client_factory()
        assert isinstance(client, websockets.client.WebSocketClientProtocol)
        assert client.subprotocol in pytest.saltyrtc.subprotocols
        yield from client.close()

    @pytest.mark.asyncio
    def test_server_ping(self, ws_client_factory):
        """
        Make sure that we can *ping* the server and the server sends a
        *pong* response.
        """
        client = yield from ws_client_factory()
        pong = yield from client.ping()
        yield from pong
        yield from client.close()

    def test_packet_min_length(self):
        """
        Check that an empty NaCl message takes exactly 40 bytes (24
        bytes nonce and 16 bytes NaCl authenticator).
        """
        a = libnacl.public.SecretKey()
        b = libnacl.public.SecretKey()
        box = libnacl.public.Box(sk=a.sk, pk=b.pk)
        nonce, data = box.encrypt(b'', pack_nonce=False)
        assert len(nonce) == 24
        assert len(data) == 16



import binascii
import os
import signal
import stat
import subprocess

import pytest

from saltyrtc.server import __version__ as _version
from saltyrtc.server import (
    Server,
    util,
)


class TestCLI:
    @pytest.mark.asyncio
    def test_invalid_command(self, cli):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli('meow')
        assert 'No such command "meow"' in exc_info.value.output

    @pytest.mark.asyncio
    def test_invalid_verbosity(self, cli):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli('-v', '8')
        assert 'is not in the valid range' in exc_info.value.output

    @pytest.mark.asyncio
    def test_import_error_logbook(self, cli, fake_logbook_env):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli('-v', '7', 'serve', '-p', '8443', env=fake_logbook_env)
        assert ('Please install saltyrtc.server[logging] for '
                'logging support') in exc_info.value.output

    @pytest.mark.asyncio
    def test_get_version(self, cli):
        output = yield from cli('-v', '7', '-c', 'version')
        assert 'Version: {}'.format(_version) in output
        assert str(Server.subprotocols) in output

    @pytest.mark.asyncio
    def test_generate_key_invalid_path(self, cli, tmpdir):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli('generate', str(tmpdir))
        assert 'is a directory' in exc_info.value.output

    @pytest.mark.asyncio
    def test_generate_key_invalid_permissions(self, cli, tmpdir):
        keyfile = tmpdir.join('keyfile.key')
        keyfile.write('meow')
        keyfile.chmod(stat.S_IREAD)
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli('generate', str(keyfile))
        assert 'is not writable' in exc_info.value.output

    @pytest.mark.asyncio
    def test_generate_key(self, cli, tmpdir):
        keyfile = tmpdir.join('keyfile.key')
        yield from cli('generate', str(keyfile))

        # Check length
        key = binascii.unhexlify(keyfile.read())
        assert len(key) == 32

        # Check permissions
        stat_result = os.stat(str(keyfile))
        permissions = stat_result.st_mode
        assert permissions & stat.S_IRWXU == stat.S_IRUSR | stat.S_IWUSR

    @pytest.mark.asyncio
    def test_serve_key_missing(self, cli):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-k', pytest.saltyrtc.permanent_key_primary,
                '-p', '8443',
            )
        assert 'It is REQUIRED' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_cert_missing(self, cli):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-sc', pytest.saltyrtc.cert,
                '-p', '8443',
            )
        assert 'It is REQUIRED' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_invalid_cert(self, cli, tmpdir):
        cert = tmpdir.join('cert.pem')
        cert.write('meowmeow')
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-sc', str(cert),
                '-k', pytest.saltyrtc.permanent_key_primary,
                '-p', '8443',
            )
        assert 'SSLError' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_invalid_key_file(self, cli, tmpdir):
        keyfile = tmpdir.join('keyfile.key')
        keyfile.write('6d656f77')
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-sc', pytest.saltyrtc.cert,
                '-k', str(keyfile),
                '-p', '8443',
            )
        assert 'ValueError' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_invalid_dh_params_file(self, cli, tmpdir):
        dh_params_file = tmpdir.join('dh_params.pem')
        dh_params_file.write('meowmeow')
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-sc', pytest.saltyrtc.cert,
                '-k', pytest.saltyrtc.permanent_key_primary,
                '-dhp', str(dh_params_file),
                '-p', '8443',
            )
        assert 'SSLError' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_invalid_hex_encoded_key(self, cli):
        key = open(pytest.saltyrtc.permanent_key_primary, 'r').read()[:63]
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-sc', pytest.saltyrtc.cert,
                '-k', key,
                '-p', '8443',
            )
        assert 'ValueError' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_invalid_host(self, cli):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-sc', pytest.saltyrtc.cert,
                '-k', pytest.saltyrtc.permanent_key_primary,
                '-h', 'meow',
                '-p', '8443',
            )
        assert 'Name or service not known' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_invalid_port(self, cli):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-sc', pytest.saltyrtc.cert,
                '-k', pytest.saltyrtc.permanent_key_primary,
                '-p', 'meow',
            )
        assert 'is not a valid integer' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_invalid_loop(self, cli):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-sc', pytest.saltyrtc.cert,
                '-k', pytest.saltyrtc.permanent_key_primary,
                '-p', '8443',
                '-l', 'meow',
            )
        assert 'invalid choice' in exc_info.value.output

    @pytest.saltyrtc.no_uvloop
    @pytest.mark.asyncio
    def test_serve_uvloop_unavailable(self, cli):
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-sc', pytest.saltyrtc.cert,
                '-k', pytest.saltyrtc.permanent_key_primary,
                '-p', '8443',
                '-l', 'uvloop',
            )
        assert "Cannot use event loop 'uvloop'" in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_asyncio(self, cli):
        output = yield from cli(
            'serve',
            '-sc', pytest.saltyrtc.cert,
            '-k', pytest.saltyrtc.permanent_key_primary,
            '-p', '8443',
            signal=signal.SIGINT,
        )
        assert 'Stopped' in output

    @pytest.mark.asyncio
    def test_serve_asyncio_dh_params(self, cli):
        output = yield from cli(
            'serve',
            '-sc', pytest.saltyrtc.cert,
            '-k', pytest.saltyrtc.permanent_key_primary,
            '-dhp', pytest.saltyrtc.dh_params,
            '-p', '8443',
            signal=signal.SIGINT,
        )
        assert 'Stopped' in output

    @pytest.mark.asyncio
    def test_serve_asyncio_hex_encoded_key(self, cli):
        output = yield from cli(
            'serve',
            '-sc', pytest.saltyrtc.cert,
            '-k', open(pytest.saltyrtc.permanent_key_primary, 'r').read(),
            '-p', '8443',
            signal=signal.SIGINT,
        )
        assert 'Stopped' in output

    @pytest.mark.asyncio
    def test_serve_asyncio_plus_logging(self, cli):
        output = yield from cli(
            '-v', '7',
            'serve',
            '-sc', pytest.saltyrtc.cert,
            '-k', pytest.saltyrtc.permanent_key_primary,
            '-p', '8443',
            signal=signal.SIGINT,
        )
        assert 'Server instance' in output
        assert 'Closing protocols' in output

    @pytest.saltyrtc.have_uvloop
    @pytest.mark.asyncio
    def test_serve_uvloop(self, cli):
        output = yield from cli(
            'serve',
            '-sc', pytest.saltyrtc.cert,
            '-k', pytest.saltyrtc.permanent_key_primary,
            '-p', '8443',
            '-l', 'uvloop',
            signal=signal.SIGINT,
        )
        assert 'Stopped' in output

    @pytest.saltyrtc.have_uvloop
    @pytest.mark.asyncio
    def test_serve_uvloop_dh_params(self, cli):
        output = yield from cli(
            'serve',
            '-sc', pytest.saltyrtc.cert,
            '-k', pytest.saltyrtc.permanent_key_primary,
            '-dhp', pytest.saltyrtc.dh_params,
            '-p', '8443',
            '-l', 'uvloop',
            signal=signal.SIGINT,
        )
        assert 'Stopped' in output

    @pytest.saltyrtc.have_uvloop
    @pytest.mark.asyncio
    def test_serve_uvloop_plus_logging(self, cli):
        output = yield from cli(
            '-v', '7',
            'serve',
            '-sc', pytest.saltyrtc.cert,
            '-k', pytest.saltyrtc.permanent_key_primary,
            '-p', '8443',
            '-l', 'uvloop',
            signal=signal.SIGINT,
        )
        assert 'Server instance' in output
        assert 'Closing protocols' in output

    @pytest.mark.asyncio
    def test_serve_asyncio_restart(self, cli):
        output = yield from cli(
            'serve',
            '-sc', pytest.saltyrtc.cert,
            '-k', pytest.saltyrtc.permanent_key_primary,
            '-p', '8443',
            signal=[signal.SIGHUP, signal.SIGINT],
        )
        output = output.split('\n')
        assert output.count('Started') == 2
        assert output.count('Stopped') == 2

    @pytest.mark.asyncio
    def test_serve_safety_not_quite_off(self, cli):
        env = os.environ.copy()
        env['SALTYRTC_SAFETY_OFF'] = 'Eh... yeah'
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                env=env,
             )
        assert 'It is REQUIRED' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_safety_off(self, cli):
        env = os.environ.copy()
        env['SALTYRTC_SAFETY_OFF'] = 'yes-and-i-know-what-im-doing'
        output = yield from cli(
            'serve',
            '-p', '8443',
            signal=signal.SIGINT,
            env=env,
        )
        assert 'Stopped' in output

    @pytest.mark.asyncio
    def test_serve_repeated_key(self, cli):
        primary_key = open(pytest.saltyrtc.permanent_key_primary, 'r').read()
        combinations = [
            ['-k', pytest.saltyrtc.permanent_key_primary,
             '-k', pytest.saltyrtc.permanent_key_primary],
            ['-k', pytest.saltyrtc.permanent_key_primary,
             '-k', primary_key],
            ['-k', pytest.saltyrtc.permanent_key_secondary,
             '-k', pytest.saltyrtc.permanent_key_secondary],
        ]

        # Try all combinations
        for key_arguments in combinations:
            with pytest.raises(subprocess.CalledProcessError) as exc_info:
                yield from cli(*[
                    'serve',
                    '-sc', pytest.saltyrtc.cert,
                    '-p', '8443',
                ] + key_arguments)
            assert 'key has been supplied more than once' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_invalid_2nd_key_file(self, cli, tmpdir):
        keyfile = tmpdir.join('keyfile.key')
        keyfile.write('6d656f77')
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-sc', pytest.saltyrtc.cert,
                '-k', pytest.saltyrtc.permanent_key_primary,
                '-k', str(keyfile),
                '-p', '8443',
            )
        assert 'ValueError' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_invalid_2nd_hex_encoded_key(self, cli):
        key = open(pytest.saltyrtc.permanent_key_primary, 'r').read()[:63]
        with pytest.raises(subprocess.CalledProcessError) as exc_info:
            yield from cli(
                'serve',
                '-sc', pytest.saltyrtc.cert,
                '-k', pytest.saltyrtc.permanent_key_primary,
                '-k', key,
                '-p', '8443',
            )
        assert 'ValueError' in exc_info.value.output

    @pytest.mark.asyncio
    def test_serve_asyncio_2nd_key(self, cli):
        # Load keys
        primary_key = util.load_permanent_key(pytest.saltyrtc.permanent_key_primary)
        primary_key = primary_key.hex_pk().decode('ascii')
        secondary_key = util.load_permanent_key(pytest.saltyrtc.permanent_key_secondary)
        secondary_key = secondary_key.hex_pk().decode('ascii')

        # Check output
        output = yield from cli(
            'serve',
            '-sc', pytest.saltyrtc.cert,
            '-k', pytest.saltyrtc.permanent_key_primary,
            '-k', pytest.saltyrtc.permanent_key_secondary,
            '-p', '8443',
            signal=signal.SIGINT,
        )
        assert 'Primary public permanent key: {}'.format(primary_key) in output
        assert 'Secondary key #1: {}'.format(secondary_key) in output
        assert 'Stopped' in output

    @pytest.mark.asyncio
    def test_serve_asyncio_2nd_key_reversed(self, cli):
        # Load keys
        primary_key = util.load_permanent_key(pytest.saltyrtc.permanent_key_primary)
        primary_key = primary_key.hex_pk().decode('ascii')
        secondary_key = util.load_permanent_key(pytest.saltyrtc.permanent_key_secondary)
        secondary_key = secondary_key.hex_pk().decode('ascii')

        # Check output
        output = yield from cli(
            'serve',
            '-sc', pytest.saltyrtc.cert,
            '-k', pytest.saltyrtc.permanent_key_secondary,
            '-k', pytest.saltyrtc.permanent_key_primary,
            '-p', '8443',
            signal=signal.SIGINT,
        )
        assert 'Primary public permanent key: {}'.format(secondary_key) in output
        assert 'Secondary key #1: {}'.format(primary_key) in output
        assert 'Stopped' in output



import asyncio
import os
import sys

import logbook.more

import saltyrtc.server


def env(name, default=None):
    return os.environ.get(name, default)


def require_env(name):
    value = env(name)
    if value is None:
        print("Missing '{}' env variable".format(name))
        sys.exit(1)
    return value


def main():
    """
    Run the SaltyRTC server until Ctrl+C has been pressed.
    """
    loop = asyncio.get_event_loop()

    # Create SSL context
    if env('SALTYRTC_DISABLE_TLS') != 'yes-and-i-know-what-im-doing':
        ssl_context = saltyrtc.server.create_ssl_context(
            certfile=require_env('SALTYRTC_TLS_CERT'),
            keyfile=require_env('SALTYRTC_TLS_KEY'),
            dh_params_file=require_env('SALTYRTC_DH_PARAMS'),
        )
    else:
        ssl_context = None

    # Get permanent key
    if env('SALTYRTC_DISABLE_SERVER_PERMANENT_KEY') != 'yes-and-i-know-what-im-doing':
        permanent_keys = [saltyrtc.server.load_permanent_key(
            require_env('SALTYRTC_SERVER_PERMANENT_KEY'))]
    else:
        permanent_keys = None

    # Start server
    port = int(env('SALTYRTC_PORT', '8765'))
    coroutine = saltyrtc.server.serve(ssl_context, permanent_keys, port=port)
    server = loop.run_until_complete(coroutine)

    # Wait until Ctrl+C has been pressed
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    # Wait until server is closed and close the event loop
    server.close()
    loop.run_until_complete(server.wait_closed())
    loop.close()


if __name__ == '__main__':
    # Enable asyncio debug logging
    os.environ['PYTHONASYNCIODEBUG'] = '1'

    # Enable logging
    saltyrtc.server.enable_logging(level=logbook.TRACE, redirect_loggers={
        'asyncio': logbook.DEBUG,
        'websockets': logbook.DEBUG,
    })

    # Run 'main'
    logging_handler = logbook.more.ColorizedStderrHandler()
    with logging_handler.applicationbound():
        main()



import ast
import os
import sys

from setuptools import setup


def get_version():
    path = os.path.join(os.path.dirname(__file__), 'saltyrtc', 'server', '__init__.py')
    with open(path) as file:
        for line in file:
            if line.startswith('__version__'):
                _, value = line.split('=', maxsplit=1)
                return ast.literal_eval(value.strip())
        else:
            raise Exception('Version not found in {}'.format(path))


def read(file):
    return open(os.path.join(os.path.dirname(__file__), file)).read().strip()


# Allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# Import long description
long_description = '\n\n'.join((read('README.rst'), read('CHANGELOG.rst')))

# Check python version
py_version = sys.version_info[:2]
if py_version < (3, 4):
    raise Exception("SaltyRTC requires Python >= 3.4")

# Logging requirements
logging_require = [
    'logbook>=1.0.0,<2',
]

# Test requirements
# Note: These are just tools that aren't required, so a version range
#       is not necessary here.
tests_require = [
    'pytest>=3.0.7',
    'pytest-asyncio>=0.5.0',
    'pytest-cov>=2.4.0',
    'flake8>=3.3.0',
    'isort>=4.2.5',
    'collective.checkdocs>=0.2',
    'Pygments>=2.2.0',  # required by checkdocs
] + logging_require

setup(
    name='saltyrtc.server',
    version=get_version(),
    packages=['saltyrtc', 'saltyrtc.server'],
    install_requires=[
        'libnacl>=1.5.0,<2',
        'click>=6.7',  # doesn't seem to follow semantic versioning
        'websockets>=3.2,<4',
        'u-msgpack-python>=2.3,<3',
    ],
    tests_require=tests_require,
    extras_require={
        ':python_version<="3.4"': [
            'asyncio>=3.4.3',
            'typing>=3,<4',
            'backports_abc==0.5',
        ],
        'dev': tests_require,
        'logging': logging_require,
        'uvloop': ['uvloop>=0.8.0,<2'],
    },
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'saltyrtc-server = saltyrtc.server.bin:main',
        ],
    },

    # PyPI metadata
    author='Lennart Grahl',
    author_email='lennart.grahl@gmail.com',
    description='A SaltyRTC compliant signalling server.',
    long_description=long_description,
    license='MIT',
    keywords='saltyrtc signalling signaling websocket websockets nacl',
    url='https://github.com/saltyrtc/saltyrtc-server-python',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Communications',
        'Topic :: Internet',
        'Topic :: Security',
    ],
)



import collections
import enum
from typing import List

try:
    from collections.abc import Coroutine
except ImportError:  # python 3.4
    from backports_abc import Coroutine

__all__ = ('Event',)


@enum.unique
class Event(enum.Enum):
    initiator_connected = 'initiator-connected'
    responder_connected = 'responder-connected'
    disconnected = 'disconnected'


class EventRegistry:
    events = collections.defaultdict(list)  # type: Dict[Event, List[Coroutine]]

    def register(self, event: Event, handler: Coroutine):
        self.events[event].append(handler)

    def get_callbacks(self, event: Event) -> List[Coroutine]:
        return self.events[event]
